@@ -34,6 +34,16 @@ class _DatasetFeatureOperand:
     feature: str
 
 
+_THRESHOLD_POSITION_MAP = {
+    ">": ThresholdPosition.ABOVE,
+    ">=": ThresholdPosition.GREATER_EQUAL,
+    "<": ThresholdPosition.BELOW,
+    "<=": ThresholdPosition.LESS_EQUAL,
+    "==": ThresholdPosition.EQUAL,
+    "!=": ThresholdPosition.NOT_EQUAL,
+}
+
+
 class _Formula:
     def __init__(
         self,
@@ -90,17 +100,15 @@ class Trigger:
     def __init__(
         self,
         name: str,
-        severity: Optional[
-            Union[
-                AlertSeverityKind,
-                Literal[
-                    "CRITICAL",
-                    "ERROR",
-                    "WARNING",
-                    "INFO",
-                ],
-            ]
-        ] = None,
+        severity: Union[
+            AlertSeverityKind,
+            Literal[
+                "CRITICAL",
+                "ERROR",
+                "WARNING",
+                "INFO",
+            ],
+        ] = AlertSeverityKind.INFO,
         threshold_position: Optional[
             Union[
                 ThresholdPosition,
@@ -205,7 +213,7 @@ P = ParamSpec("P")
 T = TypeVar("T")
 
 
-def _copy_with(function: Callable[Concatenate["Chart", P], "Chart"]) -> Callable[Concatenate["Chart", P], "Chart"]:
+def _copy_with(function: Callable[P, "Chart"]) -> Callable[P, "Chart"]:
     def inner(self: Chart, *args: P.args, **kwargs: P.kwargs) -> Chart:
         copy = deepcopy(self)
         if not self._keep:
@@ -227,23 +235,37 @@ class Chart:
     _registry: ClassVar[Set[Union[str, "Chart"]]] = set()
 
     def __init__(self, name: str, window_period: str = "1h", keep: Optional[bool] = False):
-        """
+        """Create a chart for monitoring or alerting on the Chalk dashboard.
 
         Parameters
         ----------
         name
-            The name of the chart
+            The name of the chart. If a name is not provided, the chart will be
+            named according to the series and formulas it contains.
         window_period
-            The length of the window, e.g. "20m" or "1h"
+            The length of the window, e.g. `"20m"` or `"1h"`.
+
+        Other Parameters
+        ----------------
         keep
-            By default, the builder methods that return Charts make a deepcopy
-            and be deregistered from deployment. If `keep=True`, this chart
+            By default, the builder methods that return a `Chart` make a deepcopy
+            and are not registered for deployment. If `keep=True`, this chart
             and all descendant charts will be registered automatically.
 
         Returns
         -------
         Chart
-            Your first instance of a Chart!
+            A chart for viewing in the Chalk dashboard.
+
+        Examples
+        --------
+        >>> from chalk.monitoring import Chart, Series
+        >>> Chart(name="Request count").with_trigger(
+        ...     Series
+        ...         .feature_null_ratio_metric()
+        ...         .where(feature=User.fico_score) > 0.2,
+        ... )
+
         """
         self._name = name
         self._window_period = window_period
@@ -267,24 +289,24 @@ class Chart:
         Returns
         -------
         Chart
-            A copy of your chart with the new name
+            A copy of your `Chart` with the new name.
         """
         self._name = name
         return self
 
     @_copy_with
     def with_window_period(self, window_period: str) -> "Chart":
-        """Change the window period for a chart.
+        """Change the window period for a `Chart`.
 
         Parameters
         ----------
         window_period
-            A new window period for a chart, e.g. "20m" or "1h".
+            A new window period for a chart, e.g. `"20m"` or `"1h"`.
 
         Returns
         -------
         Chart
-            A copy of your chart with the new window period
+            A copy of your `Chart` with the new window period.
         """
         parse_chalk_duration(window_period)
         self._window_period = window_period
@@ -298,7 +320,7 @@ class Chart:
         ----------
         series
             A Series instance to attach to the Chart.
-            A Chart can have any number of Series.
+            A `Chart` can have any number of `Series`.
 
         Returns
         -------
@@ -307,6 +329,12 @@ class Chart:
         """
         if not isinstance(series, SeriesBase):
             raise ValueError(f"'series' value '{series}' must be a Series object")
+        if series._entity_id and self._entity_id and series._entity_id != self._entity_id:
+            raise ValueError(
+                f"Chart '{self._name}' is already associated with '{self._entity_id}', but new Series "
+                f"'{series.name}' is associated with '{series._entity_id}'. "
+                f"Series can only be associated with one entity."
+            )
         self._series.append(series)
         if series._entity_id:
             self._entity_id = series._entity_id
@@ -321,7 +349,7 @@ class Chart:
         Parameters
         ----------
         series_name
-            The name of the Series
+            The name of the `Series`.
 
         Returns
         -------
@@ -329,13 +357,13 @@ class Chart:
             The first series added to your `Chart` with the given series name.
         """
         for series in self._series:
-            if series._name == series_name:
+            if series._name == series_name or series._default_name == series_name:
                 return series
         raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")
 
     def _get_series_index(self, series_name: str) -> Tuple[int, SeriesBase]:
         for i, series in enumerate(self._series):
-            if series._name == series_name:
+            if series._name == series_name or series._default_name == series_name:
                 return i, series
         raise ValueError(f"No series named '{series_name}' exists in Chart '{self._name}'")
 
@@ -365,18 +393,16 @@ class Chart:
     def with_trigger(
         self,
         expression: ThresholdFunction,
-        trigger_name: str,
-        severity: Optional[
-            Union[
-                AlertSeverityKind,
-                Literal[
-                    "CRITICAL",
-                    "ERROR",
-                    "WARNING",
-                    "INFO",
-                ],
-            ]
-        ] = None,
+        trigger_name: Optional[str] = None,
+        severity: Union[
+            AlertSeverityKind,
+            Literal[
+                "CRITICAL",
+                "ERROR",
+                "WARNING",
+                "INFO",
+            ],
+        ] = AlertSeverityKind.INFO,
         channel_name: Optional[str] = None,
     ) -> "Chart":
         """Attaches a `Trigger` to your `Chart`. Your `Chart` may optionally have one `Trigger`.
@@ -405,20 +431,29 @@ class Chart:
         Returns
         -------
         Chart
-            A copy of your chart with the new trigger.
+            A copy of your `Chart` with the new trigger.
         """
         if not isinstance(expression.lhs, SeriesBase):
             raise ValueError(f"Left hand side of expression '{expression.lhs}' must be a Series")
-        threshold_position = ThresholdPosition.ABOVE if expression.operation == ">" else ThresholdPosition.BELOW
-        trigger = Trigger(
+        if self._trigger is not None:
+            raise ValueError(
+                f"Trigger is already set for Chart '{self._name}'. Only one alert trigger can be set per chart."
+            )
+        if trigger_name is None:
+            trigger_name = f"{expression.lhs.name}{expression.operation}{expression.rhs}"
+
+        if expression.lhs not in self._series:
+            self._series.append(expression.lhs)
+
+        threshold_position = _THRESHOLD_POSITION_MAP[expression.operation]
+        self._trigger = Trigger(
             name=trigger_name,
             severity=severity,
             threshold_position=threshold_position,
             threshold_value=expression.rhs,
-            series_name=expression.lhs._name,
+            series_name=expression.lhs.name,
             channel_name=channel_name,
         )
-        self._trigger = trigger
         return self
 
     @_copy_with
@@ -450,7 +485,7 @@ class Chart:
         Parameters
         ----------
         resolver
-            A Chalk resolver
+            A Chalk resolver.
 
         Returns
         -------
@@ -496,7 +531,7 @@ class Chart:
         return f"Chart(name='{self._name}')"
 
     def __getitem__(self, key: str) -> Union[SeriesBase, _Formula]:
-        """Retrieve a series or formula by name from a chart
+        """Retrieve a series or formula by name from a chart.
 
         Parameters
         ----------
@@ -506,10 +541,9 @@ class Chart:
         Returns
         -------
         The series or formula with the given name.
-
         """
         for series in self._series:
-            if series._name == key:
+            if series.name == key:
                 return series
         for formula in self._formulas:
             if formula._name == key:
