@@ -1,7 +1,7 @@
 import dataclasses
 import zlib
 from copy import deepcopy
-from typing import Any, List, Literal, Optional, Tuple, TypeVar, Union
+from typing import Any, List, Literal, Optional, Tuple, TypeVar, Union, overload
 
 from chalk._monitoring.charts_enums_codegen import (
     ChartLinkKind,
@@ -11,7 +11,8 @@ from chalk._monitoring.charts_enums_codegen import (
     MetricKind,
     WindowFunctionKind,
 )
-from chalk.features.resolver import ResolverProtocol
+from chalk.features import Feature
+from chalk.features.resolver import Resolver, ResolverProtocol
 
 
 @dataclasses.dataclass
@@ -32,6 +33,23 @@ ResolverType = Literal["online", "offline", "stream"]
 TSeries = TypeVar("TSeries", bound="SeriesBase")
 
 
+_WINDOW_FUNCTION_VALUE_DICT = {
+    "99%": WindowFunctionKind.PERCENTILE_99,
+    "95%": WindowFunctionKind.PERCENTILE_95,
+    "75%": WindowFunctionKind.PERCENTILE_75,
+    "50%": WindowFunctionKind.PERCENTILE_50,
+    "25%": WindowFunctionKind.PERCENTILE_25,
+    "5%": WindowFunctionKind.PERCENTILE_5,
+    "99": WindowFunctionKind.PERCENTILE_99,
+    "95": WindowFunctionKind.PERCENTILE_95,
+    "75": WindowFunctionKind.PERCENTILE_75,
+    "50": WindowFunctionKind.PERCENTILE_50,
+    "25": WindowFunctionKind.PERCENTILE_25,
+    "5": WindowFunctionKind.PERCENTILE_5,
+    "ALL": WindowFunctionKind.ALL_PERCENTILES,
+}
+
+
 class SeriesBase:
     """
     Base class for Series
@@ -39,47 +57,61 @@ class SeriesBase:
 
     def __init__(
         self,
-        name: str,
+        name: Optional[str],
         metric: Union[MetricKind, str],
         window_function: Optional[Union[WindowFunctionKind, str]] = None,
         time_shift: Optional[int] = None,
     ):
         self._name = name
-        self._metric = MetricKind(metric.upper()) if metric else None
+        self._metric: Optional[MetricKind] = MetricKind(metric.upper()) if metric else None
         self._filters: List[MetricFilter] = []
-        if window_function:
-            if window_function.upper() in WindowFunctionKind.__members__:
-                window_function_kind = WindowFunctionKind(window_function.upper())
-            else:
-                window_function_kind = self._get_window_function_type(window_function.upper())
-                if not window_function_kind:
-                    raise ValueError(f"'window_function' value '{window_function}' 'invalid for WindowFunctionKind")
-        else:
-            window_function_kind = None
-        self._window_function = window_function_kind
+        self._window_function: Union[WindowFunctionKind, None] = self._get_window_function_type(window_function)
         self._group_by: List[GroupByKind] = []
-        self._time_shift = time_shift
+        self._time_shift: Optional[int] = time_shift
         self._entity_kind = ChartLinkKind.manual
         self._entity_id = None
+        _resolver_names = [resolver.name for resolver in Resolver.registry]
+        _resolver_fqns = [resolver.fqn for resolver in Resolver.registry]
+        self._all_valid_resolver_strings = _resolver_fqns + _resolver_names
+
+    @property
+    def _default_name(self):
+        content = ""
+        if self._window_function is not None:
+            content += (
+                self._window_function.lower().replace("all_percentiles", "pAll").replace("percentile_", "p") + ":"
+            )
+        if self._metric is not None:
+            content += self._metric.lower()
+        if len(self._group_by) > 0:
+            content += f".{'.'.join(sorted(g.value.lower() for g in self._group_by))})"
+        if len(self._filters) > 0:
+            f = ",".join(
+                sorted(
+                    f"{f.kind.value.lower()}{f.comparator.value.replace('EQ', '=').replace('NEQ', '!=').replace('ONE_OF', '∈')}{f.value[0] if len(f.value) == 1 else f.value}"
+                    for f in self._filters
+                )
+            )
+            content += f"[{f}]"
+        return content
+
+    @property
+    def name(self):
+        return self._name or self._default_name
 
     @staticmethod
-    def _get_window_function_type(key: str) -> Union[WindowFunctionKind, None]:
-        window_function_value_dict = {
-            "99%": WindowFunctionKind.PERCENTILE_99,
-            "95%": WindowFunctionKind.PERCENTILE_95,
-            "75%": WindowFunctionKind.PERCENTILE_75,
-            "50%": WindowFunctionKind.PERCENTILE_50,
-            "25%": WindowFunctionKind.PERCENTILE_25,
-            "5%": WindowFunctionKind.PERCENTILE_5,
-            "99": WindowFunctionKind.PERCENTILE_99,
-            "95": WindowFunctionKind.PERCENTILE_95,
-            "75": WindowFunctionKind.PERCENTILE_75,
-            "50": WindowFunctionKind.PERCENTILE_50,
-            "25": WindowFunctionKind.PERCENTILE_25,
-            "5": WindowFunctionKind.PERCENTILE_5,
-            "ALL": WindowFunctionKind.ALL_PERCENTILES,
-        }
-        return window_function_value_dict.get(key)
+    def _get_window_function_type(key: Optional[Union[str, WindowFunctionKind]]) -> Union[WindowFunctionKind, None]:
+        if key is None or isinstance(key, WindowFunctionKind):
+            return key
+
+        uk = key.upper()
+        if uk in WindowFunctionKind.__members__:
+            return WindowFunctionKind(uk)
+
+        if uk in _WINDOW_FUNCTION_VALUE_DICT:
+            return _WINDOW_FUNCTION_VALUE_DICT[uk]
+
+        raise ValueError(f"Window function '{key}' is not supported.")
 
     def _where(
         self: TSeries,
@@ -142,12 +174,13 @@ class SeriesBase:
 
     def _feature_name_filter(self: TSeries, *features: Tuple[Any], equals: bool) -> TSeries:
         if not features:
-            raise ValueError(f"One or more Chalk Features must be supplied.")
+            raise ValueError("One or more Chalk Features must be supplied.")
         copy = self._copy_with()
         comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
         if len(features) == 1 or not equals:
             for feature in features:
                 value = str(feature)
+                _validate_feature(value)
                 metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=comparator, value=[value])
                 copy._filters.append(metric_filter)
             if len(features) == 1:
@@ -155,19 +188,21 @@ class SeriesBase:
                 copy._entity_id = str(feature)
                 copy._entity_kind = ChartLinkKind.feature
         else:
-            value = [str(feature) for feature in features]
-            metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=ComparatorKind.ONE_OF, value=value)
+            values = [str(feature) for feature in features]
+            [_validate_feature(value) for value in values]
+            metric_filter = MetricFilter(kind=FilterKind.FEATURE_NAME, comparator=ComparatorKind.ONE_OF, value=values)
             copy._filters.append(metric_filter)
         return copy
 
     def _resolver_name_filter(self: TSeries, *resolvers: Union[ResolverProtocol, str], equals: bool) -> TSeries:
         if not resolvers:
-            raise ValueError(f"One or more Chalk Resolvers must be supplied.")
+            raise ValueError("One or more Chalk Resolvers must be supplied.")
         copy = self._copy_with()
         comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
         if len(resolvers) == 1 or not equals:
             for resolver in resolvers:
                 value = resolver if isinstance(resolver, str) else resolver.fqn
+                self._validate_resolver(value)
                 metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=comparator, value=[value])
                 copy._filters.append(metric_filter)
             if len(resolvers) == 1:
@@ -175,14 +210,15 @@ class SeriesBase:
                 copy._entity_id = resolver if isinstance(resolver, str) else resolver.fqn
                 copy._entity_kind = ChartLinkKind.resolver
         else:
-            value = [resolver if isinstance(resolver, str) else resolver.fqn for resolver in resolvers]
-            metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=ComparatorKind.ONE_OF, value=value)
+            values = [resolver if isinstance(resolver, str) else resolver.fqn for resolver in resolvers]
+            [self._validate_resolver(value) for value in values]
+            metric_filter = MetricFilter(kind=FilterKind.RESOLVER_NAME, comparator=ComparatorKind.ONE_OF, value=values)
             copy._filters.append(metric_filter)
         return copy
 
     def _string_filter(self: TSeries, *strings: str, kind: FilterKind, equals: bool = True) -> TSeries:
         if not strings:
-            raise ValueError(f"One or more arguments must be supplied for this filter")
+            raise ValueError("One or more arguments must be supplied for this filter")
         copy = self._copy_with()
         comparator = ComparatorKind.EQ if equals else ComparatorKind.NEQ
         if len(strings) == 1 or not equals:
@@ -201,7 +237,7 @@ class SeriesBase:
         self: TSeries, *resolver_types: Literal["online", "offline", "stream"], equals: bool = True
     ) -> TSeries:
         if not resolver_types:
-            raise ValueError(f"One or more resolver types from 'online', 'offline', or 'stream' must be supplied")
+            raise ValueError("One or more resolver types from 'online', 'offline', or 'stream' must be supplied")
         if not set(resolver_types).issubset(["online", "offline", "stream"]):
             raise ValueError(f"Resolver types '{resolver_types}' must be one of 'online', 'offline', or 'stream'")
         copy = self._copy_with()
@@ -221,16 +257,24 @@ class SeriesBase:
 
     def _true_false_filter(self: TSeries, kind: FilterKind, value: bool) -> TSeries:
         copy = self._copy_with()
-        value_str = "true" if value else "false"
-        metric_filter = MetricFilter(kind=kind, comparator=ComparatorKind.EQ, value=[value_str])
-        copy._filters.append(metric_filter)
+        copy._filters.append(
+            MetricFilter(
+                kind=kind,
+                comparator=ComparatorKind.EQ,
+                value=["true" if value else "false"],
+            )
+        )
         return copy
 
     def _status_filter(self: TSeries, kind: FilterKind, success: bool) -> TSeries:
         copy = self._copy_with()
-        value = "success" if success else "failure"
-        metric_filter = MetricFilter(kind=kind, comparator=ComparatorKind.EQ, value=[value])
-        copy._filters.append(metric_filter)
+        copy._filters.append(
+            MetricFilter(
+                kind=kind,
+                comparator=ComparatorKind.EQ,
+                value=["success" if success else "failure"],
+            )
+        )
         return copy
 
     def _with_filter(
@@ -267,15 +311,41 @@ class SeriesBase:
         self_copy = deepcopy(self)
         return self_copy
 
+    def _validate_resolver(self, resolver_string: str):
+        if resolver_string not in self._all_valid_resolver_strings:
+            raise ValueError(f"No resolver found with name or fqn '{resolver_string}'")
+
     def __gt__(self, other: Union[float, int]) -> ThresholdFunction:
         return ThresholdFunction(self, ">", other)
 
     def __lt__(self, other: Union[float, int]) -> ThresholdFunction:
         return ThresholdFunction(self, "<", other)
 
+    def __ge__(self, other: Union[float, int]) -> ThresholdFunction:
+        return ThresholdFunction(self, ">=", other)
+
+    def __le__(self, other: Union[float, int]) -> ThresholdFunction:
+        return ThresholdFunction(self, "<=", other)
+
+    def __ne__(self, other: Union[float, int]) -> ThresholdFunction:
+        return ThresholdFunction(self, "!=", other)
+
     def __str__(self) -> str:
         return f"Series(name='{self._name}')"
 
+    @overload
+    def __eq__(self, other: "SeriesBase") -> bool:
+        ...
+
+    @overload
+    def __eq__(self, other: Union[float, int]) -> ThresholdFunction:
+        ...
+
+    def __eq__(self, other: Union[float, int, "SeriesBase"]) -> Union[bool, ThresholdFunction]:
+        if isinstance(other, (int, float)):
+            return ThresholdFunction(self, "==", other)
+        return hash(self) == hash(other)
+
     def __hash__(self) -> int:
         name = self._name if self._name else "."
         metric = str(self._metric) if self._metric else "."
@@ -291,3 +361,12 @@ class SeriesBase:
         )
 
         return zlib.crc32(series_string.encode())
+
+
+def _validate_feature(feature_string: str):
+    split = feature_string.split(".")
+    try:
+        assert len(split) == 2
+        Feature.from_root_fqn(feature_string)
+    except Exception:
+        raise ValueError(f"No feature found with name '{feature_string}'. Feature must be in the format 'user.name'")
