@@ -43,7 +43,7 @@ def _convert_series(series: SeriesBase) -> CreateMetricConfigSeriesGQL:
     return CreateMetricConfigSeriesGQL(
         metric=convert_metric_kind(series._metric),
         filters=[_convert_filter(series_filter) for series_filter in series._filters],
-        name=series._name,
+        name=series._name or series._default_name,
         windowFunction=_convert_window_function(series._window_function),
         groupBy=[_convert_group_by(group_by) for group_by in series._group_by],
     )
