@@ -1,6 +1,6 @@
 import argparse
 import json
-import logging
+import logging as pylogging
 import os
 import sys
 import traceback
@@ -15,10 +15,10 @@ from chalk.parsed.user_types_to_json import get_registered_types_as_json
 from chalk.utils.stubgen import configure_stubgen_argparse, run_stubgen
 
 
-def _get_list_results(directory: Optional[str]):
+def _get_list_results(directory: Optional[str], file_allowlist: Optional[List[str]]):
     scope_to = Path(directory or os.getcwd())
     try:
-        failed = import_all_files()
+        failed = import_all_files(file_allowlist=file_allowlist)
         return get_registered_types_as_json(scope_to, failed)
     except Exception:
         ex_type, ex_value, ex_traceback = sys.exc_info()
@@ -29,13 +29,23 @@ def _get_list_results(directory: Optional[str]):
         return json.dumps(dict(failed=[dict(traceback=relevant_traceback)]), indent=2)
 
 
-def dump_cmd(filename: str, directory: Optional[str]):
+def dump_cmd(filename: str, directory: Optional[str], filter_file: Optional[str]):
+    file_allowlist = None
+    if filter_file is not None and os.path.exists(filter_file):
+        with open(filter_file) as f:
+            file_allowlist = [f.strip() for f in f.readlines()]
+
     with open(filename, "w") as f:
-        f.write(_get_list_results(directory))
+        f.write(
+            _get_list_results(
+                directory=directory,
+                file_allowlist=file_allowlist,
+            )
+        )
 
 
 def list_cmd(directory: str):
-    print(_get_list_results(directory))
+    print(_get_list_results(directory=directory, file_allowlist=None))
 
 
 def config_cmd():
@@ -69,12 +79,24 @@ def cli(args_override: Optional[List[str]] = None):
     dump_parser = subparsers.add_parser("dump", help="Write the resolvers for the current project to the given file")
     dump_parser.add_argument("filename", help="Write to this file")
     dump_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
+    export_parser = subparsers.add_parser(
+        "export", help="Write the resolvers for the current project to the given file"
+    )
+    export_parser.add_argument("filename", help="Write to this file")
+    export_parser.add_argument("--directory", help="Scope to this directory", nargs="?")
+    export_parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
     stubgen_parser = subparsers.add_parser("stubgen", help="Generate type stubs for feature set classes")
+    stubgen_parser.add_argument("--file_filter", help="Path containing only files to consider", nargs="?")
     configure_stubgen_argparse(stubgen_parser)
-    args = parser.parse_args(args_override)
+
+    # Parsing only known args for forwards compatibility.
+    # Changing this to `.parse_args` means once the args are
+    # set for a command, you can never add to them.
+    # Please do not change to `.parse_args`.
+    args, _ = parser.parse_known_args(args_override)
     if args.log_level:
-        level = getattr(logging, args.log_level.upper())
-        logging.basicConfig(
+        level = getattr(pylogging, args.log_level.upper())
+        pylogging.basicConfig(
             format="%(asctime)s %(levelname)s: %(message)s",
             datefmt="%m/%d/%Y %I:%M:%S %p",
             level=level,
@@ -87,13 +109,18 @@ def cli(args_override: Optional[List[str]] = None):
         config_cmd()
 
     elif args.command == "stubgen":
-        run_stubgen(args)
+        run_stubgen(args, args.file_filter)
+
+    elif args.command == "export":
+        dump_cmd(args.filename, args.directory, args.file_filter)
 
     elif args.command == "dump":
-        dump_cmd(args.filename, args.directory)
+        dump_cmd(args.filename, args.directory, None)
+
     else:
         parser.print_help(sys.stderr)
         return 1
+
     return 0
 
 
