@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import base64
+import inspect
 import json
 import os
 import uuid
@@ -16,11 +17,12 @@ import pandas as pd
 import requests as requests
 from typing_extensions import assert_never
 
-from chalk.client import ChalkClient, Dataset, DatasetRevision
+from chalk.client import ChalkBaseException, Dataset, DatasetRevision
 from chalk.client.models import (
     ChalkError,
     ColumnMetadata,
     DatasetFilter,
+    DatasetRecomputeResponse,
     DatasetResponse,
     DatasetRevisionResponse,
     OfflineQueryContext,
@@ -101,11 +103,32 @@ def _parallel_download(uris: List[str], executor: ThreadPoolExecutor, lazy: bool
         raise missing_dependency_exception("chalkpy[runtime]")
 
     df_futures: list[Future[pl.LazyFrame] | Future[pl.DataFrame]] = []
+
+    # Filesystem class registration is non-threadsafe, so let's fetch the supported filesystems here
+    # to pre-emptively register them.
+    if len(uris) > 0 and uris[0].startswith("gs"):
+        # importing here because we shouldn't assume that this is present in all cases
+        from fsspec import get_filesystem_class
+
+        get_filesystem_class("gs")
+
+    if len(uris) > 0 and uris[0].startswith("s3"):
+        # importing here because we shouldn't assume that this is present in all cases
+        from fsspec import get_filesystem_class
+
+        get_filesystem_class("s3")
+
     for uri in uris:
-        if lazy:
-            df_futures.append(executor.submit(pl.scan_parquet, uri))
-        else:
-            df_futures.append(executor.submit(pl.read_parquet, uri))
+        # New versions of polars.scan_parquet don't use fsspec, and we need to pass explicit storage options here.
+        # However, read_parquet still uses fsspec. In any case, we still collect here, so it doesn't really matter.
+        #   https://docs.rs/object_store/0.7.0/src/object_store/gcp/mod.rs.html#963-1083
+        #   https://docs.rs/object_store/0.7.0/src/object_store/aws/mod.rs.html#963-1083
+        df_futures.append(executor.submit(pl.read_parquet, uri))
+
+        # if lazy:
+        #     df_futures.append(executor.submit(pl.scan_parquet, uri))
+        # else:
+        #     df_futures.append(executor.submit(pl.read_parquet, uri))
 
     dfs = [df.result() for df in df_futures]
     dfs = [x.select(sorted(x.columns)) for x in dfs]
@@ -188,6 +211,25 @@ def _load_dataset_bigquery(
     return _extract_df_columns(df, output_feature_fqns, output_ts, output_id, version, columns)
 
 
+def to_utc(df: pl.DataFrame, col: str, expr: pl.Expr):
+    try:
+        import polars as pl
+    except ImportError:
+        raise missing_dependency_exception("chalkpy[runtime]")
+
+    if col not in df.schema:
+        return expr
+
+    dtype = df.schema[col]
+    if isinstance(dtype, pl.Datetime):
+        if dtype.time_zone is not None:
+            return expr.dt.convert_time_zone("UTC")
+        else:
+            return expr.dt.replace_time_zone("UTC")
+    else:
+        return expr
+
+
 def _extract_df_columns(
     df: Union[pl.DataFrame, pl.LazyFrame],
     output_feature_fqns: Optional[Sequence[str]],
@@ -222,7 +264,7 @@ def _extract_df_columns(
                 if c not in col_name_set:
                     ordered_cols.append(c)
             for x in column_metadata:
-                if x.feature_fqn not in ordered_cols:
+                if x.feature_fqn not in ordered_cols and x.feature_fqn in df.columns:
                     ordered_cols.append(x.feature_fqn)
             df = df.select(ordered_cols)
 
@@ -234,7 +276,9 @@ def _extract_df_columns(
             # All dataframes have an __id__ column
             expected_cols[str(ID_FEATURE)] = id_col.alias(str(ID_FEATURE))
 
-        ts_col = pl.col(str(OBSERVED_AT_FEATURE)).fill_null(pl.col(str(CHALK_TS_FEATURE))).dt.with_time_zone("UTC")
+        ts_col = to_utc(
+            df, str(OBSERVED_AT_FEATURE), pl.col(str(OBSERVED_AT_FEATURE)).fill_null(pl.col(str(CHALK_TS_FEATURE)))
+        )
         if output_ts:
             # For the ts feature, we want `__oat__` if not null else `__ts__`
             # it is not null; otherwise, we want
@@ -297,7 +341,7 @@ def _extract_df_columns(
     elif version == DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1:
         unique_features = set(df.select(pl.col("feature_name").unique()).lazy().collect()["feature_name"].to_list())
         cols = [
-            pl.col("value").filter(pl.col("feature_name") == fqn).first().alias(cast(str, fqn))
+            pl.col("value").filter(pl.col("feature_name").eq(fqn)).first().alias(cast(str, fqn))
             for fqn in unique_features
         ]
 
@@ -329,9 +373,9 @@ def _extract_df_columns(
             # and all other columns are strings
             if isinstance(dtype, pl.Datetime):
                 # Assuming that the only datetime column is for timestamps
-                decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
+                decoded_stmts.append(to_utc(df, col, pl.col(col)))
             else:
-                decoded_stmts.append(pl.col(col).apply(_json_decode))
+                decoded_stmts.append(pl.col(col).apply(_json_decode, return_dtype=dtype))
         elif version in (DatasetVersion.NATIVE_DTYPES, DatasetVersion.NATIVE_COLUMN_NAMES):
             if column_metadata is None:
                 raise ValueError(
@@ -339,7 +383,7 @@ def _extract_df_columns(
                 )
             # We already decoded the column names so matching against the fqn
             if col == CHALK_TS_FEATURE or col == OBSERVED_AT_FEATURE:
-                decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
+                decoded_stmts.append(to_utc(df, col, pl.col(col)))
             elif col == ID_FEATURE:
                 # The pkey is already decoded properly -- it's always an int or str
                 decoded_stmts.append(pl.col(col))
@@ -354,7 +398,7 @@ def _extract_df_columns(
                     )
                     if isinstance(dtype, pl.Datetime):
                         # It's probably a TS feature
-                        decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
+                        decoded_stmts.append(to_utc(df, col, pl.col(col)))
                     else:
                         decoded_stmts.append(pl.col(col))
                 else:
@@ -424,13 +468,16 @@ def load_dataset(
 
 
 class DatasetRevisionImpl(DatasetRevision):
+    _hydrated: bool
+    _show_progress: bool
+
     def __init__(
         self,
         revision_id: uuid.UUID,
         environment: EnvironmentId,
         creator_id: str,
         outputs: List[str],
-        givens_uri: str | None,
+        givens_uri: Optional[str],
         status: QueryStatus,
         filters: DatasetFilter,
         num_partitions: int,
@@ -443,7 +490,8 @@ class DatasetRevisionImpl(DatasetRevision):
         terminated_at: Optional[datetime] = None,
         dataset_name: Optional[str] = None,
         dataset_id: Optional[uuid.UUID] = None,
-        branch: Optional[str] = None,
+        branch: Optional[BranchId] = None,
+        show_progress: bool = False,
     ):
         self.revision_id = revision_id
         self.environment = environment
@@ -462,7 +510,9 @@ class DatasetRevisionImpl(DatasetRevision):
         self.dataset_name = dataset_name
         self.dataset_id = dataset_id
         self._client = client
-        self._branch: BranchId = branch
+        self._branch = BranchId(branch) if branch is not None else None
+        self._show_progress = show_progress
+        self._hydrated = self.status == QueryStatus.SUCCESSFUL
 
     @property
     def data_as_polars(self) -> pl.LazyFrame:
@@ -480,6 +530,7 @@ class DatasetRevisionImpl(DatasetRevision):
     ) -> pl.LazyFrame:
         context = OfflineQueryContext(environment=self.environment)
         _logger.info(f"loading polars LazyFrame for DatasetRevision {self.revision_id}")
+        self._hydrate(caller_method=inspect.currentframe().f_code.co_name)
         return self._client._await_offline_query_job(
             job_id=self.revision_id,
             outputs=self.outputs,
@@ -506,6 +557,7 @@ class DatasetRevisionImpl(DatasetRevision):
     ) -> pd.DataFrame:
         context = OfflineQueryContext(environment=self.environment)
         _logger.info(f"loading pandas DataFrame for DatasetRevision {self.revision_id}")
+        self._hydrate(caller_method=inspect.currentframe().f_code.co_name)
         return (
             self._client._await_offline_query_job(
                 job_id=self.revision_id,
@@ -537,6 +589,7 @@ class DatasetRevisionImpl(DatasetRevision):
     ) -> DataFrame:
         context = OfflineQueryContext(environment=self.environment)
         _logger.info(f"loading Chalk DataFrame for DatasetRevision {self.revision_id}")
+        self._hydrate(caller_method=inspect.currentframe().f_code.co_name)
         return DataFrame(
             data=self._client._await_offline_query_job(
                 job_id=self.revision_id,
@@ -556,6 +609,7 @@ class DatasetRevisionImpl(DatasetRevision):
         output_ts: bool = False,
         num_executors: Optional[int] = None,
     ) -> None:
+        self._hydrate(caller_method=inspect.currentframe().f_code.co_name)
         context = OfflineQueryContext(environment=self.environment)
         urls = self._client._await_offline_query_job(
             job_id=self.revision_id,
@@ -581,14 +635,71 @@ class DatasetRevisionImpl(DatasetRevision):
             with open(f"{directory_path}/{destination_filepath}", "wb") as f:
                 f.write(r.content)
 
-        for url in urls:
-            executor.submit(_download_data, url, path).result()
+        futures = (executor.submit(_download_data, url, path) for url in urls)
+        for f in futures:
+            f.result()
+
+    def get_input_dataframe(self) -> pl.LazyFrame:
+        context = OfflineQueryContext(environment=self.environment)
+        _logger.info(f"loading input DataFrame for DatasetRevision {self.revision_id}")
+        return self._client._await_offline_query_input(
+            job_id=self.revision_id,
+            lazy=True,
+            context=context,
+            branch=self._branch,
+        ).lazy()
 
     def __repr__(self) -> str:
         if self.dataset_name:
             return f"DatasetRevision(dataset_name='{self.dataset_name}', revision_id='{self.revision_id}', status='{self.status.value}')"
         return f"DatasetRevision(revision_id='{self.revision_id}')"
 
+    def wait_for_completion(self, show_progress: Optional[bool] = None) -> None:
+        self._hydrate(show_progress=show_progress)
+
+    def _hydrate(self, show_progress: Optional[bool] = None, caller_method: Optional[str] = None) -> None:
+        """
+        :param show_progress: Pass `True` to show a progress bar while waiting for the operation to complete.
+        :param caller_method: Caller method name. This will be used to display a user-facing message explaining
+        the implicit showing of computation progress.
+        """
+        if self._hydrated:
+            return
+
+        should_show_progress = self._show_progress
+        if show_progress is not None:
+            should_show_progress = show_progress
+
+        self._client._await_operation_completion(
+            operation_id=self.revision_id,
+            show_progress=should_show_progress,
+            caller_method=caller_method,
+            environment_id=self.environment,
+        )
+        dataset = self._client._get_anonymous_dataset(
+            revision_id=str(self.revision_id),
+            environment=self.environment,
+            branch=self._branch,
+        )
+        completed_revision = dataset.revisions[-1]
+
+        self.outputs = completed_revision.outputs
+        self.environment = completed_revision.environment
+        self.revision_id = completed_revision.revision_id
+        self._branch = completed_revision._branch
+        self.terminated_at = completed_revision.terminated_at
+        self.started_at = completed_revision.started_at
+        self.created_at = completed_revision.created_at
+        self.num_bytes = completed_revision.num_bytes
+        self.output_version = completed_revision.output_version
+        self.output_uris = completed_revision.output_uris
+        self.num_partitions = completed_revision.num_partitions
+        self.filters = completed_revision.filters
+        self.status = completed_revision.status
+        self.givens_uri = completed_revision.givens_uri
+
+        self._hydrated = True
+
 
 class DatasetImpl(Dataset):
     def __init__(
@@ -684,28 +795,61 @@ class DatasetImpl(Dataset):
             num_executors=num_executors,
         )
 
-    def recompute(self, branch: BranchId, features: Optional[List[Union[Feature, Any]]] = None):
-        new_revision = self._client._recompute_dataset(
+    def get_input_dataframe(
+        self,
+    ) -> pl.LazyFrame:
+        if len(self.revisions) == 0:
+            raise IndexError("No revisions exist for dataset")
+        return self.revisions[-1].get_input_dataframe()
+
+    def recompute(
+        self,
+        features: Optional[List[Union[Feature, Any]]] = None,
+        branch: Optional[str] = None,
+        wait: bool = False,
+        show_progress: bool = False,
+    ) -> Dataset:
+        if len(self.revisions) == 0:
+            raise IndexError("No revisions exist for dataset")
+        if self._client._config.branch is None and branch is None:
+            raise ValueError("A branch was not provided to the chalk client or as an argument to recompute.")
+
+        recompute_response = self._client._recompute_dataset(
             dataset_name=self.dataset_name,
             dataset_id=self.dataset_id,
             revision_id=self.revisions[-1].revision_id,
             features=features or [],
-            branch=branch,
+            branch=branch or self._client._config.branch,
             environment=self.environment,
         )
-        self.revisions.append(dataset_revision_from_response(new_revision, self._client))
+        if recompute_response.errors:
+            raise ChalkBaseException(errors=recompute_response.errors)
+        new_revision = dataset_revision_from_response(recompute_response, self._client)
+        new_revision._show_progress = show_progress
+        self.revisions.append(new_revision)
+
+        if not wait:
+            return self
+
+        new_revision.wait_for_completion()
         return self
 
-    def __repr__(self) -> str:
-        if self.errors:
+    def repr_help(self) -> str:
+        return "This Dataset is persisted on the server.  Call .get_data_as_*() to retrieve the data. See https://docs.chalk.ai/docs/datasets for more information."
+
+    def repr_header(self) -> str:
+        if self.errors and self.dataset_name:
             return f"Dataset(name='{self.dataset_name}', version='{self.version}', errors='{self.errors}')"
         if self.dataset_name:
             return f"Dataset(name='{self.dataset_name}', version='{self.version}')"
-        return f"Dataset()"
+        return f"Dataset(name=<unnamed>)"
+
+    def __repr__(self) -> str:
+        return "\n".join([self.repr_header(), self.repr_help()])
 
 
 def dataset_revision_from_response(
-    revision: DatasetRevisionResponse, client: ChalkAPIClientImpl
+    revision: Union[DatasetRevisionResponse, DatasetRecomputeResponse], client: ChalkAPIClientImpl
 ) -> DatasetRevisionImpl:
     return DatasetRevisionImpl(
         revision_id=revision.revision_id,
