@@ -1,14 +1,24 @@
 from __future__ import annotations
 
+import dataclasses
 import uuid
 from datetime import datetime
 from enum import Enum, IntEnum
-from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Union
+from typing import TYPE_CHECKING, Any, Dict, List, Literal, Mapping, Optional, Sequence, Union
 
 import numpy as np
 from pydantic import BaseModel, Field, root_validator
+from typing_extensions import TypeAlias
 
+from chalk.byte_transmit.model import ByteBaseModel, ByteDict
+from chalk.features import Feature, ensure_feature
+from chalk.features._encoding.json import FeatureEncodingOptions
+from chalk.features.resolver import Resolver
 from chalk.features.tag import EnvironmentId
+from chalk.utils.missing_dependency import missing_dependency_exception
+
+if TYPE_CHECKING:
+    import polars as pl
 
 
 def _category_for_error_code(c: Union[ErrorCode, str]) -> ErrorCodeCategory:
@@ -162,6 +172,9 @@ class ChalkException(BaseModel, frozen=True):
     stacktrace: str
     """The stacktrace produced by the code."""
 
+    internal_stacktrace: Optional[str] = None
+    """The stacktrace produced by the code, full detail."""
+
 
 class ChalkError(BaseModel, frozen=True):
     """
@@ -181,6 +194,16 @@ class ChalkError(BaseModel, frozen=True):
     message: str
     """A readable description of the error message."""
 
+    display_primary_key: Optional[str] = None
+    """
+    A human-readable hint that can be used to identify the entity that this error is associated with.
+    """
+
+    display_primary_key_fqn: Optional[str] = None
+    """
+    If provided, can be used to add additional context to 'display_primary_key'.
+    """
+
     exception: Optional[ChalkException] = None
     """The exception that caused the failure, if applicable."""
 
@@ -197,6 +220,9 @@ class ChalkError(BaseModel, frozen=True):
     def copy_for_feature(self, feature: str) -> "ChalkError":
         return self.copy(update={"feature": feature})
 
+    def copy_for_pkey(self, pkey: Union[str, int]) -> "ChalkError":
+        return self.copy(update={"display_primary_key": str(pkey)})
+
     @root_validator
     def _validate_category(cls, values: Dict[str, Any]):
         values["category"] = _category_for_error_code(values["code"])
@@ -275,7 +301,7 @@ class FeatureResult(BaseModel):
     this field will be empty.
     """
 
-    pkey: Optional[Union[str, int]]
+    pkey: Any = None
     """The primary key of the resolved feature."""
 
     error: Optional[ChalkError] = None
@@ -318,6 +344,37 @@ class OfflineQueryInput(BaseModel):
 class OnlineQueryRequest(BaseModel):
     inputs: Mapping[str, Any]  # Values should be of type TJSON
     outputs: List[str]
+    now: Optional[str] = None  # iso format so we can json
+    staleness: Optional[Mapping[str, str]] = None
+    context: Optional[OnlineQueryContext] = None
+    include_meta: bool = True
+    explain: Union[bool, Literal["only"]] = False
+    skip_online_storage: Optional[bool] = False
+    skip_offline_storage: Optional[bool] = False
+    skip_metrics_storage: Optional[bool] = False
+    skip_cache_lookups: Optional[bool] = False
+    correlation_id: Optional[str] = None
+    query_name: Optional[str] = None
+    deployment_id: Optional[str] = None
+    branch_id: Optional[str] = None
+    meta: Optional[Mapping[str, str]] = None
+    store_plan_stages: Optional[bool] = False
+    gcs_client: Optional[Any] = None
+    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
+
+
+@dataclasses.dataclass
+class OnlineQuery:
+    input: Union[Mapping[FeatureReference, Sequence[Any]], Any]
+    output: List[str]
+    staleness: Optional[Mapping[str, str]] = None
+    tags: Optional[List[str]] = None
+
+
+class OnlineQueryManyRequest(BaseModel):
+    inputs: Mapping[str, List[Any]]
+    outputs: List[str]
+    now: Optional[List[str]] = None
     staleness: Optional[Mapping[str, str]] = None
     context: Optional[OnlineQueryContext] = None
     include_meta: bool = True
@@ -331,6 +388,9 @@ class OnlineQueryRequest(BaseModel):
     deployment_id: Optional[str] = None
     branch_id: Optional[str] = None
     meta: Optional[Mapping[str, str]] = None
+    store_plan_stages: Optional[bool] = False
+    gcs_client: Optional[Any] = None  # deprecated
+    encoding_options: FeatureEncodingOptions = FeatureEncodingOptions()
 
 
 class TriggerResolverRunRequest(BaseModel):
@@ -396,6 +456,29 @@ class OnlineQueryResponse(BaseModel):
         }
 
 
+@dataclasses.dataclass
+class BulkOnlineQueryResult:
+    scalars_df: Optional[pl.DataFrame]
+    groups_dfs: Optional[Dict[str, pl.DataFrame]]  # change to chalk df whenever we figure out pydantic
+    errors: Optional[List[ChalkError]]
+    meta: Optional[QueryMeta]
+
+    # def get_feature_value(self, pkey: str | int, f: FeatureReference):
+    #     f_casted = ensure_feature(f)
+    #     if f_casted.is_has_many:
+    #         return self.groups_dfs[f_casted.root_fqn]
+    #     else:
+    #         return self.scalars_df[0][f_casted.root_fqn].item()
+
+
+@dataclasses.dataclass
+class BulkOnlineQueryResponse:
+    results: List[BulkOnlineQueryResult]
+
+
+FeatureReference: TypeAlias = Union[str, Any]
+
+
 class CreateOfflineQueryJobRequest(BaseModel):
     output: List[str] = Field(description="A list of output feature root fqns to query")
     required_output: List[str] = Field(default_factory=list, description="A list of required output feature root fqns")
@@ -420,16 +503,21 @@ class CreateOfflineQueryJobRequest(BaseModel):
             "Set to ``0`` to ignore the cache. If not specified, it defaults to 30 minutes."
         ),
     )
-    observed_at_lower_bound: Optional[datetime] = Field(
+    observed_at_lower_bound: Optional[str] = Field(
         default=None,
         description="The lower bound for the observed at timestamp (inclusive). If not specified, defaults to the beginning of time",
     )
-    observed_at_upper_bound: Optional[datetime] = Field(
+    observed_at_upper_bound: Optional[str] = Field(
         default=None,
         description="The upper bound for the observed at timestamp (inclusive). If not specified, defaults to the end of time.",
     )
     dataset_name: Optional[str] = None
     branch: Optional[str] = None
+    recompute_features: Union[bool, List[FeatureReference]] = False
+    store_plan_stages: bool = False
+    explain: Union[bool, Literal["onlyu"]] = False
+    tags: Optional[List[str]] = None
+    correlation_id: Optional[str] = None
 
 
 class ComputeResolverOutputRequest(BaseModel):
@@ -442,8 +530,8 @@ class ComputeResolverOutputRequest(BaseModel):
 class DatasetRecomputeRequest(BaseModel):
     dataset_name: Optional[str] = None
     branch: str
-    dataset_id: Optional[str]
-    revision_id: Optional[str]
+    dataset_id: Optional[str] = None
+    revision_id: Optional[str] = None
     features: List[str]
 
 
@@ -552,7 +640,7 @@ class DatasetRevisionResponse(BaseModel):
     dataset_name: Optional[str] = None
     dataset_id: Optional[uuid.UUID] = None
     environment_id: EnvironmentId
-    revision_id: uuid.UUID  # Currently, the revision ID is the job ID that created the revision
+    revision_id: Optional[uuid.UUID] = None  # Currently, the revision ID is the job ID that created the revision
     creator_id: str
     outputs: List[str]
     created_at: Optional[datetime] = None
@@ -568,6 +656,35 @@ class DatasetRevisionResponse(BaseModel):
     branch: Optional[str] = None
 
 
+class DatasetRecomputeResponse(DatasetRevisionResponse):
+    errors: Optional[List[ChalkError]] = None
+
+    @classmethod
+    def from_revision_response(
+        cls, revision: DatasetRevisionResponse, errors: Optional[List[ChalkError]] = None
+    ) -> "DatasetRecomputeResponse":
+        return cls(
+            revision_id=revision.revision_id,
+            environment=revision.environment_id,
+            creator_id=revision.creator_id,
+            outputs=revision.outputs,
+            givens_uri=revision.givens_uri,
+            status=revision.status,
+            filters=revision.filters,
+            num_partitions=revision.num_partitions,
+            output_uris=revision.output_uris,
+            output_version=revision.output_version,
+            num_bytes=revision.num_bytes,
+            created_at=revision.created_at,
+            started_at=revision.started_at,
+            terminated_at=revision.terminated_at,
+            dataset_name=revision.dataset_name,
+            dataset_id=revision.dataset_id,
+            branch=revision.branch,
+            errors=errors,
+        )
+
+
 class DatasetResponse(BaseModel):
     is_finished: bool = Field(description="Whether the export job is finished (it runs asynchronously)")
     version: int = Field(
@@ -584,9 +701,247 @@ class DatasetResponse(BaseModel):
     errors: Optional[List[ChalkError]] = None
 
 
+class SingleEntityUpdate(BaseModel):
+    entity_type: Literal["feature", "resolver"]
+    entity_fqn: str
+    entity_shortname: str
+
+    @classmethod
+    def for_resolver(cls, resolver: Resolver) -> "SingleEntityUpdate":
+        return cls(
+            entity_type="resolver",
+            entity_fqn=resolver.fqn,
+            entity_shortname=resolver.fqn.split(".")[-1],
+        )
+
+    @classmethod
+    def for_feature(cls, feature: Feature) -> "SingleEntityUpdate":
+        return cls(
+            entity_type="feature",
+            entity_fqn=feature.fqn,
+            entity_shortname=feature.name,
+        )
+
+
+class UpdateGraphEntityResponse(BaseModel):
+    """
+    Represents the result of live updating a graph entity like a resolver or feature class.
+    This may result in multiple individual resolvers/features being updated, e.g. if the user
+    adds a new feature class w/ multiple new fields.
+    """
+
+    added: Optional[List[SingleEntityUpdate]] = None
+    modified: Optional[List[SingleEntityUpdate]] = None
+    removed: Optional[List[SingleEntityUpdate]] = None
+
+    errors: Optional[List[ChalkError]] = None
+
+
 class UpdateResolverResponse(BaseModel):
     updated_fqn: Optional[
         str
     ] = None  # The resolver fqn that was updated (may not be the same as the one that was requested)
     is_new: Optional[bool] = None  # Whether a new resolver was created, or if an existing one was replaced
     errors: Optional[List[ChalkError]] = None
+
+
+class FeatureObservationDeletionRequest(BaseModel):
+    """
+    Represents a request to target particular feature observations for deletion. Note that
+    the "features" and "tags" fields are mutually exclusive -- either only one of them is
+    specified, or neither is specified, in which case deletion will proceed for all
+    features of the primary keys specified.
+    """
+
+    namespace: str
+    """
+    The namespace in which the features targeted for deletion reside.
+    """
+
+    features: Optional[List[str]]
+    """
+    An optional list of the feature names of the features that should be deleted
+    for the targeted primary keys. Not specifying this and not specifying the "tags" field
+    will result in all features being targeted for deletion for the specified primary keys.
+    Note that this parameter and the "tags" parameter are mutually exclusive.
+    """
+
+    tags: Optional[List[str]]
+    """
+    An optional list of tags that specify features that should be targeted for deletion.
+    If a feature has a tag in this list, its observations for the primary keys you listed
+    will be targeted for deletion. Not specifying this and not specifying the "features"
+    field will result in all features being targeted for deletion for the specified primary
+    keys. Note that this parameter and the "features" parameter are mutually exclusive.
+    """
+
+    primary_keys: List[str]
+    """
+    The primary keys of the observations that should be targeted for deletion.
+    """
+
+
+class FeatureObservationDeletionResponse(BaseModel):
+    """
+    Contains ChalkErrors for any failures, if any, that might have occurred when trying
+    to delete the features that were requested.
+    """
+
+    errors: Optional[List[ChalkError]]
+
+
+class FeatureDropRequest(BaseModel):
+    namespace: str
+    """Namespace in which the features targeted for drop reside."""
+
+    features: List[str]
+    """Names of the features that should be dropped."""
+
+
+class FeatureDropResponse(BaseModel):
+    """
+    Contains ChalkErrors for any failures, if any, that might have occurred when trying
+    to drop the features that were requested.
+    """
+
+    errors: Optional[List[ChalkError]]
+
+
+class GetIncrementalProgressResponse(BaseModel):
+    """
+    Returns information about the current state of an incremental resolver.
+    Specifically, the recorded timestamps that the resolver uses to process recent data.
+    If both timestamp fields are returned as None, this means the current resolver hasn't
+    run yet or hasn't stored any progress data. The next time it runs it will ingest all historical data
+
+    More information at https://docs.chalk.ai/docs/sql#incremental-queries
+    """
+
+    environment_id: EnvironmentId
+
+    resolver_fqn: str
+    """
+    The fully qualified name of the given resolver
+    """
+
+    max_ingested_timestamp: Optional[datetime]
+    """
+    The latest timestamp found in ingested data.
+    """
+
+    last_execution_timestamp: Optional[datetime]
+    """
+    The latest timestamp at which the resolver was run. If configured to do so, the
+    resolver uses this timestamp instead of max_ingested_timestamp to filter input data.
+    If None, this means that this value isn't currently used by this resolver.
+    """
+
+    errors: Optional[List[ChalkError]] = None
+
+
+class SetIncrementalProgressRequest(BaseModel):
+    """
+    Sets the current state of an incremental resolver, specifically the timestamps it uses
+    to filter inputs to only recent data, to the given timestamps.
+
+    More information at https://docs.chalk.ai/docs/sql#incremental-queries
+    """
+
+    max_ingested_timestamp: Optional[datetime] = None
+    """
+    The latest timestamp found in ingested data.
+    Timestamp must have a timezone specified.
+    """
+
+    last_execution_timestamp: Optional[datetime] = None
+    """
+    The latest time the resolver was run. If configured to do so, the
+    resolver uses this timestamp instead of max_ingested_timestamp to filter input data.
+    Timestamp must have a timezone specified.
+    """
+
+
+class BranchDeployRequest(BaseModel):
+    branch_name: str
+    """
+    Name of the branch. If branch does not exist, it will be created.
+    """
+
+    create_only: bool = False
+    """
+    If true, tries to create a new branch returns an error if the branch already exists.
+    """
+
+    source_deployment_id: Optional[str] = None
+    """
+    Use the given deployment's source on the branch. If None, the latest active deployment will be used.
+    """
+
+
+class BranchDeployResponse(BaseModel):
+    branch_name: str
+    new_branch_created: bool
+
+    source_deployment_id: str
+    branch_deployment_id: str
+
+
+BranchIdParam: TypeAlias = Union[None, str, type(Ellipsis)]
+"""
+Type used for the 'branch' paremeter in calls to the Chalk Client.
+The branch can either be:
+ 1. A string that is used as the branch name for the request
+ 2. None, in which case the request is _not_ sent to a branch server,
+ 3. Ellipsis (...), indicating that the branch name (or lack thereof) is
+    inferred from the ChalkClient's current branch.
+"""
+
+
+class StreamResolverTestStatus(str, Enum):
+    SUCCESS = "SUCCESS"
+    FAILURE = "FAILURE"
+
+
+class StreamResolverTestMessagePayload(BaseModel):
+    key: str
+    message_str: Optional[str]
+    message_bytes: Optional[str]
+    timestamp: Optional[datetime]
+
+
+class StreamResolverTestRequest(BaseModel):
+    resolver_fqn: str
+    num_messages: Optional[int] = None
+    test_messages: Optional[List[StreamResolverTestMessagePayload]] = None
+
+
+class StreamResolverTestResponse(BaseModel):
+    status: StreamResolverTestStatus
+    data_uri: Optional[str] = None
+    errors: Optional[List[ChalkError]] = None
+    message: Optional[str] = None
+
+    @property
+    def features(self) -> pl.DataFrame:
+        if self.data_uri is None:
+            raise ValueError(
+                "Features were not saved to storage. "
+                "Please inspect 'ResolverTestResponse.errors' and 'ResolverTestResponse.message'."
+            )
+        try:
+            import polars as pl
+        except ImportError:
+            raise missing_dependency_exception("chalkpy[runtime]")
+        return pl.read_parquet(self.data_uri)
+
+
+class OnlineQueryResultFeather(ByteBaseModel):
+    has_data: bool
+    scalar_data: bytes
+    groups_data: ByteDict
+    errors: Optional[List[str]]  # inner str is json of ChalkError to mimic Optional[List[ChalkError]]
+    meta: Optional[str]  # inner str is json of QueryMeta to mimic Optional[QueryMeta]
+
+
+class OnlineQueryResponseFeather(ByteBaseModel):
+    query_results_bytes: ByteDict
