@@ -16,6 +16,7 @@ from chalk.features.hooks import after_all, before_all
 from chalk.features.primary import Primary, is_primary
 from chalk.features.resolver import Cron, Resolver, ResolverProtocol, offline, online, sink
 from chalk.features.tag import Environments, Tags
+from chalk.features.underscore import Underscore, _, underscore
 from chalk.utils import MachineType
 
 
@@ -158,6 +159,7 @@ __all__ = [
     "MissingValueStrategy",
     "Primary",
     "ResolverProtocol",
+    "Underscore",
     "deserialize_dtype",
     "TPrimitive",
     "Tags",
@@ -175,6 +177,7 @@ __all__ = [
     "has_one",
     "is_feature_time",
     "is_primary",
+    "is_features_cls",
     "offline",
     "online",
     "op",
@@ -184,4 +187,6 @@ __all__ = [
     "sink",
     "tags",
     "unwrap_feature",
+    "_",
+    "underscore",
 ]
