@@ -6,7 +6,12 @@ from typing import TYPE_CHECKING, Any, Generic, Optional, Protocol, Sequence, Ty
 
 import pyarrow as pa
 
-from chalk.features._encoding.json import structure_json_to_primitive, unstructure_primitive_to_json
+from chalk.features._encoding.json import (
+    FeatureEncodingOptions,
+    _structs_as_objects_feature_json_converter,
+    structure_json_to_primitive,
+    unstructure_primitive_to_json,
+)
 from chalk.features._encoding.missing_value import MissingValueStrategy
 from chalk.features._encoding.primitive import TPrimitive
 from chalk.features._encoding.pyarrow import pyarrow_to_polars, pyarrow_to_primitive, rich_to_pyarrow
@@ -50,10 +55,10 @@ class FeatureConverter(Generic[_TPrim, _TRich]):
         self,
         name: str,
         is_nullable: bool,
-        rich_default: Union[_TRich, ellipsis] = ...,
-        primitive_default: Union[_TPrim, ellipsis] = ...,
-        rich_type: Union[Type[_TRich], ellipsis] = ...,
-        pyarrow_dtype: Optional[pa.DataType] = None,
+        rich_default: Union[_TRich, Ellipsis] = ...,
+        primitive_default: Union[_TPrim, Ellipsis] = ...,
+        rich_type: Union[Type[_TRich], Ellipsis] = ...,
+        pyarrow_dtype: Optional[pa.DataType[Any]] = None,
         encoder: Optional[TEncoder[_TPrim, _TRich]] = None,
         decoder: Optional[TDecoder[_TPrim, _TRich]] = None,
     ) -> None:
@@ -181,9 +186,7 @@ class FeatureConverter(Generic[_TPrim, _TRich]):
         # For example, if a string is passed in for a datetime value, convert it into a datetime
         if self.is_value_missing(value):
             if missing_value_strategy == "allow":
-                warnings.warn(
-                    UserWarning(f"Allowing missing value for feature '{self._name}' with strategy 'default_or_allow'")
-                )
+                warnings.warn(UserWarning(f"Allowing missing value for feature '{self._name}' with strategy 'allow'"))
                 return cast(_TPrim, value)
             elif missing_value_strategy in ("default_or_error", "default_or_allow"):
                 if self.has_default:
@@ -217,15 +220,18 @@ class FeatureConverter(Generic[_TPrim, _TRich]):
         self,
         value: Union[_TRich, ellipsis, None],
         missing_value_strategy: MissingValueStrategy = "default_or_allow",
+        options: FeatureEncodingOptions = FeatureEncodingOptions(),
     ) -> TJSON:
         prim_val = self.from_rich_to_primitive(value, missing_value_strategy)
-        return self.from_primitive_to_json(prim_val)
+        return self.from_primitive_to_json(prim_val, options=options)
 
     def from_pyarrow_to_rich(self, values: Union[pa.Array, pa.ChunkedArray], /) -> Sequence[_TRich]:
         return [self.from_primitive_to_rich(x) for x in values.to_pylist()]
 
-    def from_pyarrow_to_json(self, values: Union[pa.Array, pa.ChunkedArray]) -> Sequence[TJSON]:
-        return [self.from_primitive_to_json(x) for x in self.from_pyarrow_to_primitive(values)]
+    def from_pyarrow_to_json(
+        self, values: Union[pa.Array, pa.ChunkedArray], options: FeatureEncodingOptions = FeatureEncodingOptions()
+    ) -> Sequence[TJSON]:
+        return [self.from_primitive_to_json(x, options=options) for x in self.from_pyarrow_to_primitive(values)]
 
     def from_pyarrow_to_primitive(self, values: Union[pa.Array, pa.ChunkedArray]) -> Sequence[_TPrim]:
         return values.to_pylist()
@@ -237,8 +243,13 @@ class FeatureConverter(Generic[_TPrim, _TRich]):
         x = pa.array([None if x is ... else x for x in value], type=self._pyarrow_dtype)
         return x
 
-    def from_primitive_to_json(self, value: TPrimitive) -> TJSON:
-        return unstructure_primitive_to_json(value)
+    def from_primitive_to_json(
+        self, value: TPrimitive, options: FeatureEncodingOptions = FeatureEncodingOptions()
+    ) -> TJSON:
+        if options.encode_structs_as_objects:
+            return _structs_as_objects_feature_json_converter.unstructure_primitive_to_json(value)
+        else:
+            return unstructure_primitive_to_json(value)
 
     def from_json_to_rich(self, value: TJSON) -> _TRich:
         prim_val = self.from_json_to_primitive(value)
@@ -303,3 +314,12 @@ class FeatureConverter(Generic[_TPrim, _TRich]):
                 return False
             return not self._is_nullable
         return False
+
+    def is_rich_valid(self, value: _TRich):
+        """Returns true if value has a valid rich type"""
+        try:
+            prim = self.from_rich_to_primitive(value, "default_or_error")
+            pa.scalar(prim, type=self.pyarrow_dtype)
+            return True
+        except (TypeError, ValueError):
+            return False
