@@ -5,6 +5,8 @@ import collections.abc
 import dataclasses
 import decimal
 import enum
+import ipaddress
+import uuid
 from datetime import date, datetime, time, timedelta
 from typing import Any, FrozenSet, List, Set, Tuple, Type, TypeVar, Union, cast
 
@@ -12,10 +14,12 @@ import attrs
 import cattrs
 import dateutil.parser
 import isodate
+import pydantic
 from pydantic import BaseModel
-from typing_extensions import get_args, get_origin, get_type_hints, is_typeddict
+from typing_extensions import get_args, get_origin, is_typeddict
 
 from chalk.features._encoding.primitive import TPrimitive
+from chalk.utils.cached_type_hints import cached_get_type_hints
 from chalk.utils.collections import is_namedtuple, unwrap_optional_and_annotated_if_needed
 from chalk.utils.enum import get_enum_value_type
 
@@ -68,7 +72,7 @@ def _structure_struct(obj: Any, typ: Type):
         # to ensure that any recursive fields are properly structured
         obj = _rich_converter.unstructure(obj)
     typ = unwrap_optional_and_annotated_if_needed(typ)
-    type_hints = get_type_hints(typ)
+    type_hints = cached_get_type_hints(typ)
     if obj is None:
         obj = {k: None for k in type_hints}
     if isinstance(obj, collections.abc.Sequence):
@@ -299,6 +303,67 @@ def _structure_datetime(x: Any, typ: Type[datetime]) -> datetime:
 _rich_converter.register_structure_hook(datetime, _structure_datetime)
 
 
+##########
+# UUID
+##########
+
+_rich_converter.register_unstructure_hook(uuid.UUID, lambda x: x)
+
+
+def _structure_uuid(x: Any, typ: Type[uuid.UUID]) -> uuid.UUID:
+    if x is None:
+        return cast(uuid.UUID, None)
+    if isinstance(x, uuid.UUID):
+        return x
+    if isinstance(x, str):
+        return uuid.UUID(x)
+    raise TypeError(f"Cannot structure {x} into a uuid.UUID")
+
+
+_rich_converter.register_structure_hook(uuid.UUID, _structure_uuid)
+
+
+"""
+ipaddress.IPv4Network
+"""
+_rich_converter.register_unstructure_hook(ipaddress.IPv4Address, lambda x: x)
+
+
+def _structure_ipv4(x: Any, typ: Type[ipaddress.IPv4Address]) -> ipaddress.IPv4Address:
+    if x is None:
+        return cast(ipaddress.IPv4Address, None)
+    if isinstance(x, ipaddress.IPv4Address):
+        return x
+    if isinstance(x, str):
+        return ipaddress.IPv4Address(x)
+    if isinstance(x, int):
+        return ipaddress.IPv4Address(x)
+    raise TypeError(f"Cannot structure {x} into an ipaddress.IPv4Address")
+
+
+_rich_converter.register_structure_hook(ipaddress.IPv4Address, _structure_ipv4)
+
+"""
+ipaddress.IPv6Network
+"""
+_rich_converter.register_unstructure_hook(ipaddress.IPv6Address, lambda x: x)
+
+
+def _structure_ipv6(x: Any, typ: Type[ipaddress.IPv6Address]) -> ipaddress.IPv6Address:
+    if x is None:
+        return cast(ipaddress.IPv6Address, None)
+    if isinstance(x, ipaddress.IPv6Address):
+        return x
+    if isinstance(x, str):
+        return ipaddress.IPv6Address(x)
+    if isinstance(x, int):
+        return ipaddress.IPv6Address(x)
+    raise TypeError(f"Cannot structure {x} into an ipaddress.IPv6Address")
+
+
+_rich_converter.register_structure_hook(ipaddress.IPv6Address, _structure_ipv6)
+
+
 #############
 # Numpy types
 #############
@@ -333,10 +398,19 @@ def _structure_rich(obj: Any, typ: Type):
                 return False
             if obj == 1:
                 return True
+        if np and isinstance(obj, np.bool_):
+            return bool(obj)
         raise ValueError(
             f"Cannot convert {obj} to a boolean. Allowed values are y, yes, t, true, 1, n, no, f, false, 0."
         )
 
+    if (
+        isinstance(obj, (list, set, frozenset, dict, tuple, pydantic.BaseModel))
+        or dataclasses.is_dataclass(obj)
+        or attrs.has(type(obj))
+    ):
+        raise ValueError(f"Object {obj} of type {type(obj).__name__} cannot be converted into a {typ.__name__}")
+
     return typ(obj)
 
 
@@ -348,8 +422,30 @@ _rich_converter.register_structure_hook(bool, _structure_rich)
 
 
 def unstructure_rich_to_primitive(val: Any) -> TPrimitive:
+    if val is None or type(val) in (str, int, float, bytes, time, datetime, date, timedelta, bool):
+        # Short-circuit these rich types, as the primitive types are the same
+        return val
+    # Special case for enums, particularly enums that use a mixin
+    # i.e. class MyEnum(str, enum.Enum)
+    # cattrs seems unable to detect that the class inherits from Enum so doesn't call the correct serde hooks
+    if isinstance(val, enum.Enum):
+        return _rich_converter.unstructure(val.value)
     return _rich_converter.unstructure(val)
 
 
 def structure_primitive_to_rich(val: TPrimitive, typ: Type[TRich]) -> TRich:
+    if (val is None or type(val) is typ) and typ in (
+        str,
+        int,
+        bool,
+        float,
+        bytes,
+        timedelta,
+        decimal.Decimal,
+        date,
+        datetime,
+        time,
+    ):
+        # Short circuit these rich types
+        return cast(TRich, val)
     return _rich_converter.structure(val, typ)
