@@ -35,13 +35,27 @@ class _DTypeCode(enum.IntEnum):
     DATE32 = 17
     DATE64 = 18
     DURATION = 19  # time unit
+    """UNSUPPORTED -- Some database backends do not support duration columns."""
+
     BINARY = 20  # length
+    """DEPRECATED -- use LARGE_BINARY
+
+    BINARY does not support more than 2GB of data, which significantly limits the maximum chunk size.
+    """
     STRING = 21
+    """Deprecated -- use LARGE_STRING
+
+    STRING does not support more than 2GB of data, which significantly limits the maximum chunk size.
+    """
     LARGE_BINARY = 22
     LARGE_STRING = 23
     # Decimal is not yet supported by chalk
     # DECIMAL128 = 24  # precision, scale
     LIST = 25  # value type, size
+    """Deprecated -- use LARGE_LIST
+
+    LIST does not support more than 2GB of data, which significantly limits the maximum chunk size.
+    """
     LARGE_LIST = 26  # value type
     STRUCT = 27  # fields
 
@@ -87,6 +101,9 @@ class _SingletonDType(_BaseSerializedDType):
     def to_pyarrow_dtype(self) -> pa.DataType:
         if self.type_code == _DTypeCode.BOOL:
             return pa.bool_()
+        if self.type_code == _DTypeCode.STRING:
+            # String is deprecated; automatically replacing with LargeString
+            return pa.large_utf8()
         try:
             return getattr(pa, self.type_code.name.lower())()
         except AttributeError:
@@ -115,7 +132,9 @@ class _BinaryDType(_BaseSerializedDType):
     length: int
 
     def to_pyarrow_dtype(self) -> pa.DataType:
-        return pa.binary(self.length)
+        # Binary is deprecated; automatically upgrading to LargeBinary
+        return pa.large_binary()
+        # return pa.binary(self.length)
 
 
 class _ListDType(_BaseSerializedDType):
@@ -124,7 +143,9 @@ class _ListDType(_BaseSerializedDType):
     length: int
 
     def to_pyarrow_dtype(self) -> pa.DataType:
-        return pa.list_(self.inner_dtype.to_pyarrow_dtype(), self.length)
+        # List is deprecated; automatically upgrading to LargeList
+        return pa.large_list(self.inner_dtype.to_pyarrow_dtype())
+        # return pa.list_(self.inner_dtype.to_pyarrow_dtype(), self.length)
 
 
 class _LargeListDType(_BaseSerializedDType):
@@ -148,7 +169,7 @@ class _StructDType(_BaseSerializedDType):
     type_code: Literal[_DTypeCode.STRUCT]
     fields: List[_StructField]
 
-    def to_pyarrow_dtype(self) -> pa.Field:
+    def to_pyarrow_dtype(self) -> pa.DataType:
         return pa.struct([x.to_pyarrow_dtype() for x in self.fields])
 
 
@@ -207,12 +228,15 @@ def _serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
     if pa.types.is_float64(pa_dtype):
         return _SingletonDType(type_code=_DTypeCode.FLOAT64)
     if pa.types.is_time32(pa_dtype):
+        assert isinstance(pa_dtype, pa.Time32Type)
         return _TimeUnitDType(type_code=_DTypeCode.TIME32, time_unit=_TimeUnitCode(pa_dtype.unit))
 
     if pa.types.is_time64(pa_dtype):
+        assert isinstance(pa_dtype, pa.Time64Type)
         return _TimeUnitDType(type_code=_DTypeCode.TIME64, time_unit=_TimeUnitCode(pa_dtype.unit))
 
     if pa.types.is_timestamp(pa_dtype):
+        assert isinstance(pa_dtype, pa.TimestampType)
         return _TimestampDType(
             type_code=_DTypeCode.TIMESTAMP,
             time_unit=_TimeUnitCode(pa_dtype.unit),
@@ -223,25 +247,31 @@ def _serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
         return _SingletonDType(type_code=_DTypeCode.DATE32)
     if pa.types.is_date64(pa_dtype):
         return _SingletonDType(type_code=_DTypeCode.DATE64)
-    if pa.types.is_duration(pa_dtype):
-        return _TimeUnitDType(type_code=_DTypeCode.DURATION, time_unit=_TimeUnitCode(pa_dtype.unit))
+    # if pa.types.is_duration(pa_dtype):
+    #     assert isinstance(pa_dtype, pa.DurationType)
+    #     return _TimeUnitDType(type_code=_DTypeCode.DURATION, time_unit=_TimeUnitCode(pa_dtype.unit))
 
     if pa.types.is_large_string(pa_dtype):
         return _SingletonDType(type_code=_DTypeCode.LARGE_STRING)
     if pa.types.is_large_binary(pa_dtype):
         return _SingletonDType(type_code=_DTypeCode.LARGE_BINARY)
     if pa.types.is_fixed_size_binary(pa_dtype):
+        assert isinstance(pa_dtype, pa.FixedSizeBinaryType)
         return _BinaryDType(type_code=_DTypeCode.BINARY, length=pa_dtype.byte_width)
     if pa.types.is_binary(pa_dtype):
-        return _BinaryDType(type_code=_DTypeCode.BINARY, length=-1)
+        raise ValueError("The `Binary` data type is not supported. Instead, use `LargeBinary`")
+    #     return _BinaryDType(type_code=_DTypeCode.BINARY, length=-1)
     if pa.types.is_string(pa_dtype):
-        return _SingletonDType(type_code=_DTypeCode.STRING)
+        raise ValueError("The `String` data type is not supported. Instead, use `LargeString`")
+    #     return _SingletonDType(type_code=_DTypeCode.STRING)
 
     if pa.types.is_large_list(pa_dtype):
+        assert isinstance(pa_dtype, pa.LargeListType)
         return _LargeListDType(
             type_code=_DTypeCode.LARGE_LIST, inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type)
         )
     if pa.types.is_fixed_size_list(pa_dtype):
+        assert isinstance(pa_dtype, pa.FixedSizeListType)
         return _ListDType(
             type_code=_DTypeCode.LIST,
             inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type),
@@ -249,6 +279,7 @@ def _serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
         )
 
     if pa.types.is_list(pa_dtype):
+        assert isinstance(pa_dtype, pa.ListType)
         return _ListDType(
             type_code=_DTypeCode.LIST,
             inner_dtype=_serialize_pyarrow_dtype(pa_dtype.value_type),
@@ -256,6 +287,7 @@ def _serialize_pyarrow_dtype(pa_dtype: pa.DataType) -> _SerializedDType:
         )
 
     if pa.types.is_struct(pa_dtype):
+        assert isinstance(pa_dtype, pa.StructType)
         schema = pa.schema(pa_dtype)
         return _StructDType(
             type_code=_DTypeCode.STRUCT,
