@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import pyarrow as pa
 import ast
 import copy
 import functools
@@ -13,16 +14,20 @@ from datetime import datetime, timedelta
 from itertools import islice
 from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union, cast, overload
 
+from chalk.features.filter import Filter
 from chalk.features._class_property import classproperty, classproperty_support
 from chalk.features.feature_field import Feature, _VersionInfo
 from chalk.features.feature_set import Features, FeatureSetBase
 from chalk.features.feature_time import feature_time
 from chalk.features.feature_wrapper import FeatureWrapper, unwrap_feature
+from chalk.features.underscore import Underscore
 from chalk.features.tag import Tags
 from chalk.serialization.parsed_annotation import ParsedAnnotation
 from chalk.streams import Windowed
+from chalk.utils import notebook
 from chalk.utils.collections import ensure_tuple
 from chalk.utils.duration import Duration, parse_chalk_duration
+from chalk.utils.environment_parsing import env_var_bool
 from chalk.utils.metaprogramming import MISSING, create_fn, field_assign, set_new_attribute
 from chalk.utils.string import removeprefix, to_snake_case
 
@@ -41,6 +46,7 @@ def features(
     etl_offline_to_online: bool = False,
     max_staleness: Optional[Duration] = None,
     name: Optional[str] = None,
+    singleton: bool = False,
 ) -> Callable[[Type[T]], Type[T]]:
     ...
 
@@ -58,6 +64,7 @@ def features(
     etl_offline_to_online: bool = False,
     max_staleness: Optional[Duration] = None,
     name: Optional[str] = None,
+    singleton: bool = False,
 ) -> Union[Callable[[Type[T]], Type[T]], Type[T]]:
     """Chalk lets you spell out your features directly in Python.
 
@@ -108,7 +115,10 @@ def features(
     ... )
     ... class User:
     ...     id: str
+    ...     # Comments here appear in the web!
+    ...     # :tags: pii
     ...     name: str | None
+    ...     # :owner: userteam@mycompany.com
     ...     location: LatLng
     """
 
@@ -135,12 +145,10 @@ def features(
             etl_offline_to_online=etl_offline_to_online,
             max_staleness=max_staleness,
             namespace=namespace,
+            singleton=singleton,
         )
         assert issubclass(updated_class, Features)
-        FeatureSetBase.registry[updated_class.__chalk_namespace__] = updated_class
-        if FeatureSetBase.hook is not None:
-            FeatureSetBase.hook(cast(Features, updated_class))
-
+        _add_features_class_to_registry(updated_class)
         return cast(Type[T], updated_class)
 
     # See if we're being called as @features or @features().
@@ -152,11 +160,49 @@ def features(
     return wrap(cls)
 
 
+def _add_features_class_to_registry(new_features_class: Features):
+    previous_features_class = FeatureSetBase.registry.get(new_features_class.__chalk_namespace__, None)
+    if previous_features_class is not None:
+        if notebook.is_notebook() or env_var_bool("CHALK_ALLOW_REGISTRY_UPDATES"):
+            new_feature_fqns = {fqn for fqn in new_features_class.__chalk_feature_fqns_raw__}
+            previous_feature_fqns = {fqn for fqn in previous_features_class.__chalk_feature_fqns_raw__}
+            missing_fqns = previous_feature_fqns - new_feature_fqns
+            if len(missing_fqns) > 0:
+                raise ValueError(
+                    f"New feature class definitions must include existing features. The missing features are: {sorted(missing_fqns)}"
+                )
+        else:
+            raise ValueError(
+                f"Feature class '{new_features_class.__name__}' is defined twice: "
+                f"once in '{new_features_class.__module__}' and once in '{previous_features_class.__module__}'."
+            )
+
+    FeatureSetBase.registry[new_features_class.__chalk_namespace__] = new_features_class
+
+    if new_features_class.__chalk_is_singleton__:
+        FeatureSetBase.__chalk_singletons__[new_features_class.__chalk_namespace__] = new_features_class
+
+    if FeatureSetBase.hook is not None:
+        FeatureSetBase.hook(cast(Features, new_features_class))
+
+
 def _discover_feature(
-    features: Sequence[Feature], name: str, *conditions: Callable[[Feature], bool]
+    fs: Sequence[Feature],
+    name: str,
+    *conditions: Callable[[Feature], bool],
 ) -> Optional[Feature]:
+    """
+    Parameters
+    ----------
+    fs
+        The features to search
+    name
+        Used for error messages
+    conditions
+        Tested in order. The first feature that matches _any_ condition is returned.
+    """
     for cond in conditions:
-        filtered_features = [c for c in features if cond(c)]
+        filtered_features = [c for c in fs if cond(c)]
         if len(filtered_features) == 1:
             return filtered_features[0]
         if len(filtered_features) > 1:
@@ -272,6 +318,15 @@ def _parse_tags(ts: str) -> List[str]:
     return [xx.strip() for xx in ts.split(" ")]
 
 
+def _parse_windows(ws: str) -> tuple[int, ...]:
+    windows = re.sub(" *,? *", " ", ws).strip().split(" ")
+    return tuple(int(parse_chalk_duration(w).total_seconds()) for w in windows)
+
+
+def _get_windowed_pseudofeature_name(name: str, bucket: Union[str, int]) -> str:
+    return f"{name}__{bucket}__"
+
+
 def _get_field(
     cls: Type,
     annotation_name: str,
@@ -281,6 +336,7 @@ def _get_field(
     class_etl_offline_to_online: bool,
     class_max_staleness: timedelta,
     namespace: str,
+    is_singleton: bool,
 ) -> Feature:
     # Return a Field object for this field name and type.  ClassVars and
     # InitVars are also returned, but marked as such (see f._field_type).
@@ -311,11 +367,21 @@ def _get_field(
         # Convert it to a Feature
         f = default._to_feature(bucket=None)
     else:
+        underscore_expression = None
         # The feature was not set explicitly
         if isinstance(default, types.MemberDescriptorType):
             # This is a field in __slots__, so it has no default value.
             default = ...
-        f = Feature(name=annotation_name, namespace=namespace, default=default)
+        if isinstance(default, Underscore):
+            underscore_expression = default
+            default = ...
+
+        f = Feature(
+            name=annotation_name,
+            namespace=namespace,
+            default=default,
+            underscore_expression=underscore_expression,
+        )
 
     # Only at this point do we know the name and the type.  Set them.
     f.namespace = namespace
@@ -323,7 +389,15 @@ def _get_field(
 
     f.features_cls = cls
     f.attribute_name = annotation_name
-    _process_field(f, comments, class_owner, class_tags, class_etl_offline_to_online, class_max_staleness)
+    f.is_singleton = is_singleton
+    _process_field(
+        f=f,
+        comments=comments,
+        class_owner=class_owner,
+        class_tags=class_tags,
+        class_etl_offline_to_online=class_etl_offline_to_online,
+        class_max_staleness=class_max_staleness,
+    )
     return f
 
 
@@ -339,6 +413,7 @@ def _process_field(
     comment_based_description = None
     comment_based_owner = None
     comment_based_tags = None
+    comment_based_windows = None
 
     if comment_for_feature is not None:
         comment_lines = []
@@ -353,6 +428,13 @@ def _process_field(
                         comment_based_tags = parsed
                     else:
                         comment_based_tags.extend(parsed)
+            elif stripped_line.startswith(":windows:"):
+                parsed = _parse_windows(removeprefix(stripped_line, ":windows:"))
+                if len(parsed) > 0:
+                    if comment_based_windows is None:
+                        comment_based_windows = parsed
+                    else:
+                        comment_based_windows = comment_based_windows + parsed
             else:
                 comment_lines.append(line)
 
@@ -361,6 +443,13 @@ def _process_field(
     if f.description is None and comment_based_description is not None:
         f.description = comment_based_description
 
+    # TODO: Add support for comment-based windows
+    # if comment_based_windows is not None:
+    #     if len(f.window_durations) == 0:
+    #         f.window_durations = tuple(comment_based_windows)
+    #     else:
+    #         f.window_durations = tuple(f.window_durations) + tuple(comment_based_windows)
+
     if comment_based_tags is not None:
         if f.tags is None:
             f.tags = comment_based_tags
@@ -375,10 +464,8 @@ def _process_field(
 
     if f.owner is not None and comment_based_owner is not None:
         raise ValueError(
-            (
-                f"Owner for {f.namespace}.{f.name} specified both on the feature and in the comment. "
-                f"Please use only one of these two."
-            )
+            f"Owner for {f.namespace}.{f.name} specified both on the feature and in the comment. "
+            f"Please use only one of these two."
         )
     elif f.owner is None:
         f.owner = comment_based_owner or class_owner
@@ -420,7 +507,7 @@ def _repr_fn(_globals: Dict[str, Any]):
         args=["self"],
         body=[
             'return f"{self.__class__.__qualname__}(" + '
-            + f"', '.join(f'{{x[0]}}={{x[1]}}' for x in {tuples} if x[1] != MISSING)"
+            + f"', '.join(f'{{x[0]}}={{x[1]}}' for x in {tuples} if x[1] is not MISSING)"
             + '+ ")"'
         ],
         _globals=_globals,
@@ -452,7 +539,7 @@ def _len_fn(_globals: Dict[str, Any]):
         args=["self"],
         body=[
             "__chalk_count__ = 0",
-            f"for __chalk_f__ in self.features:",
+            "for __chalk_f__ in self.features:",
             "    if not __chalk_f__.no_display and hasattr(self, __chalk_f__.attribute_name):",
             "        __chalk_count__ += 1",
             "return __chalk_count__",
@@ -467,7 +554,7 @@ def _items_fn(_globals: Dict[str, Any]):
         name="items",
         args=["self"],
         body=[
-            f"for __chalk_f__ in self.features:",
+            "for __chalk_f__ in self.features:",
             "    if not __chalk_f__.no_display and hasattr(self, __chalk_f__.attribute_name):",
             "        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
         ],
@@ -481,15 +568,22 @@ def _iter_fn(_globals: Mapping[str, Any]):
         "__iter__",
         args=["self"],
         body=[
-            f"for __chalk_f__ in self.features:",
-            f"    if hasattr(self, __chalk_f__.attribute_name) and type(__chalk_f__).__name__ == 'Feature' and not __chalk_f__.is_has_one and not __chalk_f__.is_has_many and not __chalk_f__.no_display:",
-            f"        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
+            "for __chalk_f__ in self.features:",
+            "    if hasattr(self, __chalk_f__.attribute_name) and type(__chalk_f__).__name__ == 'Feature' and not __chalk_f__.is_has_one and not __chalk_f__.is_has_many and not __chalk_f__.no_display:",
+            "        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
         ],
     )
 
 
-def _parse_annotation_comments(cls: Type) -> Mapping[str, str]:
-    source = textwrap.dedent(inspect.getsource(cls))
+def _get_class_source(cls: Type) -> Optional[str]:
+    if notebook.is_defined_in_module(cls):
+        return inspect.getsource(cls)
+    # TODO (rkargon) Try to get source from notebook cells if possible
+    return None
+
+
+def _parse_annotation_comments(class_source: str, cls_annotations: Dict[str, Any]) -> Mapping[str, str]:
+    source = textwrap.dedent(class_source)
 
     source_lines = source.splitlines()
     tree = ast.parse(source)
@@ -511,9 +605,18 @@ def _parse_annotation_comments(cls: Type) -> Mapping[str, str]:
                 if len(comments) > 0:
                     comments_for_annotations[stmt.target.id] = textwrap.dedent("\n".join(comments))
 
+    """Attach the same comments to windowed features as well"""
+    for annotation, feature_type in cls_annotations.items():
+        if annotation in comments_for_annotations and isinstance(feature_type, Windowed):
+            for bucket_size in feature_type.buckets_seconds:
+                pseudofeature_name = _get_windowed_pseudofeature_name(annotation, bucket_size)
+                comments_for_annotations[pseudofeature_name] = comments_for_annotations[annotation]
     return comments_for_annotations
 
 
+CHALK_SINGLETON_VALUE = 111
+
+
 def _process_class(
     cls: Type[T],
     owner: Optional[str],
@@ -521,11 +624,19 @@ def _process_class(
     etl_offline_to_online: bool,
     max_staleness: timedelta,
     namespace: str,
+    singleton: bool,
 ) -> Type[T]:
     raw_cls_annotations = cls.__dict__.get("__annotations__", {})
 
+    alias_from_to: Dict[str, str] = {}
+    additional_inits: list[tuple[str, Any]] = []
+
     cls_annotations: Dict[str, Any] = {}
     for name, annotation in raw_cls_annotations.items():
+        if name in ["features", "namespace", "items"]:
+            raise ValueError(
+                f"Feature name '{name}' in '{namespace}' is reserved for Chalk metafeatures. Please rename."
+            )
         if isinstance(annotation, Windowed):
             # NOTE: For Windowed resolvers, both the Annotation and the value are instances of Windowed,
             # unlike normal features whose annotation is the underlying type, and the value is an instance
@@ -548,10 +659,36 @@ def _process_class(
             for bucket in wind._buckets:
                 # Make pseudofeatures for each bucket of the window
                 feat = wind._to_feature(bucket=bucket)
-                setattr(cls, feat.name, feat)
-                # For the psuedofeatures, which track an individual bucket, the correct annotation is the
-                # underlying annotation, not Windowed[underlying], since it's only one value
+                # For the pseudofeatures, which track an individual bucket,
+                # the correct annotation is the underlying annotation, not
+                # Windowed[underlying], since it's only one value
                 cls_annotations[feat.name] = wind.kind
+                setattr(cls, feat.name, feat)
+                alias = f"{name}_{bucket}"
+                if bucket != "all":
+                    bucket = str(int(parse_chalk_duration(bucket).total_seconds()))
+                additional_inits.append((alias, annotation._kind))
+                alias_from_to[_get_windowed_pseudofeature_name(name, bucket)] = alias
+                # days = b // 86400
+                # days_as_seconds = days * 86400
+                # hours = (b - days_as_seconds) // 3600
+                # hours_as_seconds = hours * 3600
+                # minutes = (b - days_as_seconds - hours_as_seconds) // 60
+                # minutes_as_seconds = minutes * 60
+                # seconds = b - days_as_seconds - hours_as_seconds - minutes_as_seconds
+                #
+                # alias = f"{name}_" + "".join(
+                #     [
+                #         f"{count}{unit}"
+                #         for count, unit in (
+                #             (days, "d"),
+                #             (hours, "h"),
+                #             (minutes, "m"),
+                #             (seconds, "s"),
+                #         )
+                #         if count > 0
+                #     ]
+                # )
 
         cls_annotations[name] = annotation
 
@@ -571,18 +708,51 @@ def _process_class(
                     f"Feature '{namespace}.{name}' is missing an annotation. Please add one, like '{name}: str = ...'"
                 )
             cls_annotations[name] = member.typ.annotation
+        # catch malformed features like num_transactions: int = windowed("2h") with no Windowed type
+        if isinstance(member, Windowed) and not isinstance(cls_annotations.get(name), Windowed):
+            raise TypeError(
+                f"Feature '{namespace}.{name}' is marked as 'windowed()', "
+                f"but also needs to be marked as a Windowed type. "
+                f"Please add one, like '{name}: Windowed[int] = ...'"
+            )
 
     cls.__annotations__ = cls_annotations
 
-    # If we pass this function something that isn't a class, it could raise
     try:
-        comments = _parse_annotation_comments(cls)
+        class_source = _get_class_source(cls)
     except:
-        comments = {}
+        class_source = None
+    set_new_attribute(cls=cls, name="__chalk_class_definition__", value=class_source)
+    set_new_attribute(cls=cls, name="__chalk_is_singleton__", value=singleton)
 
     cls_fields: List[Feature] = []
 
-    def __chalk_primary__(cls: Type[Features]):
+    if singleton:
+        for c in cls_fields:
+            if c.primary:
+                raise ValueError(
+                    f"The singleton feature class '{namespace}' includes a feature '{c.name}' that is primary. "
+                    f"Please remove the feature '{c.fqn}' "
+                    f"or remove the singleton keyword argument to the feature class '{namespace}'."
+                )
+
+        f = Feature(
+            primary=True,
+            attribute_name="__chalk_singleton_id__",
+            name="__chalk_singleton_id__",
+            default=CHALK_SINGLETON_VALUE,
+            namespace=namespace,
+            typ=int,
+            pyarrow_dtype=pa.uint8(),
+            max_staleness=None,
+            etl_offline_to_online=False,
+            is_autogenerated=True,
+            no_display=True,
+        )
+        f.is_singleton = True
+        cls_fields.append(f)
+
+    def __chalk_primary__(_: Type[Features]):
         return _discover_feature(
             cls_fields,
             "primary",
@@ -595,7 +765,7 @@ def _process_class(
 
     set_new_attribute(cls, "__chalk_primary__", value=classproperty(__chalk_primary__, cached=True))
 
-    def __chalk_ts__(cls: Type[Features]) -> Feature:
+    def __chalk_ts__(cl: Type[Features]) -> Feature:
         # Not using `f.is_feature_time` as that would create an infinite recursion, since
         # `.is_feature_time` accesses `__chalk_ts__`
         ts_feature: Optional[Feature] = _discover_feature(
@@ -606,10 +776,14 @@ def _process_class(
             lambda f: f.name == "ts" and not f.has_resolved_join and not f._primary and not f.typ.is_primary,
         )
         if ts_feature is None:
-            return unwrap_feature(getattr(cls, GENERATED_OBSERVED_AT_NAME))
+            return unwrap_feature(getattr(cl, GENERATED_OBSERVED_AT_NAME))
         return ts_feature
 
-    set_new_attribute(cls=cls, name="__chalk_ts__", value=classproperty(__chalk_ts__, cached=True))
+    set_new_attribute(
+        cls=cls,
+        name="__chalk_ts__",
+        value=classproperty(__chalk_ts__, cached=True),
+    )
 
     def __chalk_observed_at__(cls: Type[Features]) -> FeatureWrapper:
         ts_feature: Optional[Feature] = _discover_feature(
@@ -634,7 +808,7 @@ def _process_class(
         ts_feature.is_autogenerated = True
         cls.__annotations__[GENERATED_OBSERVED_AT_NAME] = datetime
         _process_field(
-            ts_feature,
+            f=ts_feature,
             comments={},
             class_owner=cls.__chalk_owner__,
             class_tags=tuple(cls.__chalk_tags__),
@@ -650,14 +824,29 @@ def _process_class(
         value=classproperty(__chalk_observed_at__, cached=True, bind_to_instances=False),
     )
 
-    def __features__(cls: Features) -> List[Feature]:
-        features = list(cls_fields)
-        if cls.__chalk_ts__ not in features:
-            assert cls.__chalk_ts__ is not None
-            features.append(cls.__chalk_ts__)
-        return features
+    def __features__(cl: Features) -> List[Feature]:
+        fs = list(cls_fields)
+        if cl.__chalk_ts__ not in fs:
+            assert cl.__chalk_ts__ is not None
+            fs.append(cl.__chalk_ts__)
+        return fs
+
+    def __features_raw__(_: Features) -> List[Feature]:
+        return cls_fields
+
+    def __feature_fqns_raw__(_: Features) -> List[str]:
+        """
+        List of explicit feature FQN's defined as members in the class.
+        Used for validation; in certain contexts we can't call __features__
+        since it visits the feature definitions (e.g. join keys) which might contain forward references
+        """
+        return [x.fqn for x in cls_fields]
 
     set_new_attribute(cls=cls, name="features", value=classproperty(__features__, cached=True))
+    set_new_attribute(
+        cls=cls, name="__chalk_feature_fqns_raw__", value=classproperty(__feature_fqns_raw__, cached=True)
+    )
+    set_new_attribute(cls=cls, name="__chalk_features_raw__", value=classproperty(__features_raw__, cached=True))
     set_new_attribute(cls=cls, name="__repr__", value=_repr_fn(_globals=_globals))
     set_new_attribute(cls=cls, name="__eq__", value=_eq_fn(_globals=_globals))
     set_new_attribute(cls=cls, name="__hash__", value=None)
@@ -680,6 +869,12 @@ def _process_class(
     # Moving this line lower causes all kinds of problems.
     cls = classproperty_support(cls)
 
+    try:
+        # If we pass this function something that isn't a class, it could raise
+        comments = _parse_annotation_comments(class_source, cls_annotations)
+    except:
+        comments = {}
+
     # Parse the fields after we have the correct `cls` set
     cls_fields.extend(
         _get_field(
@@ -691,13 +886,10 @@ def _process_class(
             class_etl_offline_to_online=etl_offline_to_online,
             class_max_staleness=max_staleness,
             namespace=namespace,
+            is_singleton=singleton,
         )
         for name in cls_annotations
     )
-
-    alias_from_to: Dict[str, str] = {}
-
-    # We're appending to this list
     for f in islice(cls_fields, len(cls_fields)):
         if f.version is None:
             continue
@@ -742,7 +934,7 @@ def _process_class(
         cls=cls,
         name="__init__",
         value=_init_fn(
-            feature_names_and_annotations=[(x.attribute_name, x.typ.annotation) for x in cls_fields],
+            feature_names_and_annotations=[(x.attribute_name, x.typ.annotation) for x in cls_fields] + additional_inits,
             alias_from_to=alias_from_to,
             _globals=_globals,
         ),
@@ -750,6 +942,7 @@ def _process_class(
 
     for f in cls_fields:
         assert f.attribute_name is not None
+        f.features_cls = cls
         # Wrap all class features with FeatureWrapper
         setattr(cls, f.attribute_name, FeatureWrapper(f))
 
