@@ -97,7 +97,7 @@ class FeatureWrapper:
         return self._cmp("or", other)
 
     def __repr__(self):
-        return f"FeatureWrapper(fqn={self._chalk_feature.namespace}.{self._chalk_feature.name}, typ={self._chalk_feature.typ})"
+        return f"FeatureWrapper(fqn={self._chalk_feature.root_fqn}, typ={self._chalk_feature.typ})"
 
     def __str__(self):
         return str(self._chalk_feature)
@@ -143,8 +143,13 @@ class FeatureWrapper:
         assert isinstance(self._chalk_feature.typ.parsed_annotation, type)
         if self._chalk_feature.typ is not None and issubclass(self._chalk_feature.typ.parsed_annotation, DataFrame):
             f_copy = FeatureWrapper(copy.copy(self._chalk_feature))
-
             f_copy._chalk_feature.typ = ParsedAnnotation(underlying=self._chalk_feature.typ.parsed_annotation[item])
+            if (
+                hasattr(f_copy._chalk_feature, "path")
+                and f_copy._chalk_feature.path is not None
+                and len(f_copy._chalk_feature.path) > 0
+            ):
+                f_copy._chalk_feature.path[-1].child = unwrap_feature(f_copy)
 
             return f_copy
 
