@@ -1,10 +1,11 @@
 import inspect
 import logging
-from typing import Any, Callable, Iterable, List, Optional, Set, Union, overload
+from typing import Any, Callable, Iterable, Optional, Set, Tuple, Union, overload
 
 from typing_extensions import TypeAlias
 
 from chalk.features.tag import Environments
+from chalk.utils.collections import ensure_tuple
 
 HookFn: TypeAlias = Callable[[], Any]
 
@@ -24,20 +25,14 @@ class Hook:
     before_all: Set["Hook"] = set()
     after_all: Set["Hook"] = set()
 
-    environment: Optional[List[str]]
+    environment: Optional[Tuple[str, ...]]
     fn: HookFn
     filename: str
 
     def __init__(self, fn: HookFn, filename: str, environment: Optional[Environments] = None):
         self.fn = fn
         self.filename = filename
-        self.environment = (
-            [
-                environment,
-            ]
-            if isinstance(environment, str)
-            else environment
-        )
+        self.environment = None if environment is None else ensure_tuple(environment)
 
     def __call__(self):
         return self.fn()
@@ -60,16 +55,16 @@ def before_all(fn: HookFn, /) -> Hook:
 
 
 @overload
-def before_all(fn: None = None, /, environment: Environments = None) -> Callable[[HookFn], Hook]:
+def before_all(fn: None = None, /, environment: Optional[Environments] = None) -> Callable[[HookFn], Hook]:
     ...
 
 
 def before_all(
-    fn: Optional[HookFn] = None, /, environment: Environments = None
+    fn: Optional[HookFn] = None, /, environment: Optional[Environments] = None
 ) -> Union[Hook, Callable[[HookFn], Hook]]:
     caller_filename = inspect.stack()[1].filename
 
-    def decorator(f):
+    def decorator(f: HookFn):
         hook = Hook(fn=f, filename=caller_filename, environment=environment)
         Hook.before_all.add(hook)
         return hook
@@ -78,21 +73,21 @@ def before_all(
 
 
 @overload
-def after_all(fn: HookFn, /, environment: Environments = None) -> Hook:
+def after_all(fn: HookFn, /, environment: Optional[Environments] = None) -> Hook:
     ...
 
 
 @overload
-def after_all(fn: None = None, /, environment: Environments = None) -> Callable[[HookFn], Hook]:
+def after_all(fn: None = None, /, environment: Optional[Environments] = None) -> Callable[[HookFn], Hook]:
     ...
 
 
 def after_all(
-    fn: Optional[HookFn] = None, /, environment: Environments = None
+    fn: Optional[HookFn] = None, /, environment: Optional[Environments] = None
 ) -> Union[Hook, Callable[[HookFn], Hook]]:
     caller_filename = inspect.stack()[1].filename
 
-    def decorator(f):
+    def decorator(f: HookFn):
         hook = Hook(fn=f, filename=caller_filename, environment=environment)
         Hook.after_all.add(hook)
         return hook
