@@ -5,6 +5,7 @@ from chalk.features.feature_field import Feature
 
 __all__ = [
     "CHALK_TS_FEATURE",
+    "Now",
     "ID_FEATURE",
     "OBSERVED_AT_FEATURE",
     "REPLACED_OBSERVED_AT_FEATURE",
@@ -20,6 +21,13 @@ CHALK_TS_FEATURE = Feature(
     max_staleness=None,
     etl_offline_to_online=False,
 )
+Now = Feature(
+    name="now",
+    namespace=PSEUDONAMESPACE,
+    typ=datetime,
+    max_staleness=None,
+    etl_offline_to_online=False,
+)
 ID_FEATURE = Feature(
     name="__id__",
     namespace=PSEUDONAMESPACE,
@@ -42,4 +50,10 @@ REPLACED_OBSERVED_AT_FEATURE = Feature(
     etl_offline_to_online=False,
 )
 
-PSEUDOFEATURES: List[Feature] = [CHALK_TS_FEATURE, ID_FEATURE, OBSERVED_AT_FEATURE, REPLACED_OBSERVED_AT_FEATURE]
+PSEUDOFEATURES: List[Feature] = [
+    Now,
+    CHALK_TS_FEATURE,
+    ID_FEATURE,
+    OBSERVED_AT_FEATURE,
+    REPLACED_OBSERVED_AT_FEATURE,
+]
