@@ -12,4 +12,13 @@ class _UserLoggerFilter(logging.Filter):
 # Named to "chalk.clogging.chalk_logger" for backwards compatibility
 # The name must begin with `chalk` so it will be picked up by the logging filters
 chalk_logger = chalk.utils.log_with_context.get_logger("chalk.clogging.chalk_logger")
+"""A logger for use in resolvers.
+
+Examples
+--------
+>>> @online
+... def fn() -> User.name:
+...     chalk_logger.info("running")
+...     return ...
+"""
 chalk_logger.addFilter(_UserLoggerFilter())
