@@ -1,6 +1,8 @@
+from __future__ import annotations
+
 import collections
 from collections import defaultdict
-from typing import Dict, List, Mapping
+from typing import DefaultDict, Dict, List, Mapping
 
 from chalk.parsed.duplicate_input_gql import (
     GraphLogSeverity,
@@ -10,6 +12,7 @@ from chalk.parsed.duplicate_input_gql import (
     UpsertGraphGQL,
     UpsertResolverGQL,
 )
+from chalk.utils import string
 from chalk.utils.duration import parse_chalk_duration
 from chalk.utils.string import oxford_comma_list, to_camel_case
 
@@ -38,11 +41,7 @@ def _validate_primary_key(feature: UpsertFeatureGQL, builder: _LogBuilder):
     if (
         feature.scalarKind is not None
         and feature.scalarKind.primary
-        and feature.scalarKind.scalarKind
-        not in [
-            "str",
-            "int",
-        ]
+        and feature.scalarKind.scalarKind not in ["str", "int"]
     ):
         builder.add_error(
             header=f'Invalid primary key type "{feature.id.fqn}"',
@@ -97,7 +96,7 @@ def _validate_no_feature_times_as_input(
 
 
 def _validate_feature_names_unique(features: List[UpsertFeatureGQL], builder: _LogBuilder):
-    counter: dict[str, int] = defaultdict(lambda: 0)
+    counter: DefaultDict[str, int] = defaultdict(lambda: 0)
     for r in features:
         counter[r.id.fqn] += 1
 
@@ -113,7 +112,7 @@ def _validate_feature_names_unique(features: List[UpsertFeatureGQL], builder: _L
 
 
 def _validate_resolver_names_unique(resolvers: List[UpsertResolverGQL], builder: _LogBuilder):
-    counter: dict[str, int] = defaultdict(lambda: 0)
+    counter: DefaultDict[str, int] = defaultdict(lambda: 0)
     for r in resolvers:
         counter[r.fqn.split(".")[-1]] += 1
 
@@ -140,26 +139,32 @@ def _validate_offline_resolver_has_one(resolver: UpsertResolverGQL, builder: _Lo
             full_path.append(i.path[-1].parentToChildAttributeName)
             pathname = ".".join(full_path)
             builder.add_error(
-                header=f"Invalid offline resolver input",
-                subheader=f"Offline resolvers cannot require nested of more than one level. "
-                + f'The resolver "{resolver.fqn}" takes as input the feature "{pathname}"',
+                header="Invalid offline resolver input",
+                subheader=(
+                    "Offline resolvers cannot require nested of more than one level. "
+                    f'The resolver "{resolver.fqn}" takes as input the feature "{pathname}".'
+                ),
             )
 
 
-def _validate_resolver_input(resolver: UpsertResolverGQL, builder: _LogBuilder):
+def _validate_resolver_input(
+    singleton_namespaces: set[str],
+    resolver: UpsertResolverGQL,
+    builder: _LogBuilder,
+):
     namespaces = set()
 
     for i in resolver.inputs or []:
-        if i.path is not None and len(i.path) > 0:
-            namespaces.add(i.path[0].parent.namespace)
-        else:
-            namespaces.add(i.underlying.namespace)
+        f = i.path[0].parent if i.path is not None and len(i.path) > 0 else i.underlying
+        if f.namespace not in singleton_namespaces:
+            namespaces.add(f.namespace)
+
     namespaces = list(namespaces)
     if len(namespaces) > 1:
         builder.add_error(
             header=f'Resolver "{resolver.fqn}" requires features from multiple namespaces.',
             subheader=f"""All inputs to a resolver should be rooted in the same namespace.
-The resolver "{resolver.fqn}" takes inputs in the namespaces {oxford_comma_list(namespaces)}.
+The resolver "{resolver.fqn}" takes inputs in the namespaces {oxford_comma_list(string.add_quotes(namespaces))}.
 
 If you require features from many feature classes, reference them via their relationships, such as:
   User.bank_account.title,
@@ -199,18 +204,21 @@ def _validate_resolver_output(resolver: UpsertResolverGQL, builder: _LogBuilder)
     #         subheader="See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.",
     #     )
 
-    if len(features) > 0:
-        namespaces = list(set(f.namespace for f in features))
-        if len(namespaces) > 1:
-            builder.add_error(
-                header=f'Resolver "{resolver.fqn}" outputs features from multiple namespaces.',
-                subheader=f"""All outputs of a resolver should be rooted in the same namespace.
-The resolver "{resolver.fqn}" outputs features in the namespaces {oxford_comma_list(namespaces)}.
-See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.
+    # Bring this back.
 
-{resolver.functionDefinition}
-""",
-            )
+
+#     if len(features) > 0:
+#         namespaces = list(set(f.namespace for f in features))
+#         if len(namespaces) > 1:
+#             builder.add_error(
+#                 header=f'Resolver "{resolver.fqn}" outputs features from multiple namespaces.',
+#                 subheader=f"""All outputs of a resolver should be rooted in the same namespace.
+# The resolver "{resolver.fqn}" outputs features in the namespaces {oxford_comma_list(namespaces)}.
+# See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.
+#
+# {resolver.functionDefinition}
+# """,
+#             )
 
 
 # FIXME CHA-66 we should validate that joins are all pkey on pkey joins. no non-primary keys, no self-joins, no constants
@@ -227,7 +235,7 @@ def _validate_feature_names(feature: UpsertFeatureGQL, builder: _LogBuilder):
                 'Feature names cannot begin with "_chalk" or "__". Please rename this feature.'
             ),
         )
-    if feature.id.namespace.startswith("__") or feature.id.name.startswith("_chalk"):
+    if feature.id.namespace.startswith("__") or feature.id.namespace.startswith("_chalk"):
         builder.add_error(
             header="Feature class uses protected namespace",
             subheader=(
@@ -237,9 +245,58 @@ def _validate_feature_names(feature: UpsertFeatureGQL, builder: _LogBuilder):
         )
 
 
-# FIXME: This needs to check for cycles. In particular:
-#   def resolve_x(id: User.id) -> DataFrame[User]
-# is bogus
+def _validate_resolver_feature_cycles(
+    fqn_to_feature: Mapping[str, UpsertFeatureGQL],
+    resolver: UpsertResolverGQL,
+    builder: _LogBuilder,
+):
+    input_root_fqns: set[str] = set()
+    for inp in resolver.inputs or []:
+        if fqn_to_feature[inp.underlying.fqn].scalarKind and not fqn_to_feature[inp.underlying.fqn].scalarKind.primary:
+            if not inp.path:
+                input_root_fqns.add(inp.underlying.fqn)
+            else:
+                path_parsed: List[str] = []
+                path_parsed.append(inp.path[0].parent.namespace)
+                path_parsed.append(inp.path[0].parent.name)
+                for x in inp.path:
+                    path_parsed.append(x.child.name)
+                input_root_fqns.add(".".join(path_parsed))
+        # FIXME: Validate cycles for has-one and has-many features
+
+    output_features = list(resolver.output.features or [])
+    output_features += [f for df in resolver.output.dataframes for f in (df.columns or [])]
+    output_fqns: set[str] = set()
+    for f in output_features:
+        fqn = f"{f.namespace}.{f.name}"
+        root_fqn = f.fqn  # for output features, the fqn is the root fqn
+        if fqn_to_feature[fqn].scalarKind is not None and not fqn_to_feature[fqn].scalarKind.primary:
+            output_fqns.add(root_fqn)
+
+    shared_features = output_fqns & input_root_fqns
+    if len(shared_features) > 0:
+        builder.add_error(
+            header=f'Resolver "{resolver.fqn}" has the same feature in its input and output (i.e. a cycle).',
+            subheader=f"""The inputs and outputs of a resolver can't contain the same exact feature.
+The resolver "{resolver.fqn}" contains the following shared features in its inputs and outputs: {sorted(shared_features)}
+""",
+        )
+
+
+def _validate_resolver_input_and_output_namespace(resolver: UpsertResolverGQL, builder: _LogBuilder):
+    input_features = [i.underlying for i in (resolver.inputs or [])]
+    output_features = resolver.output.features or []
+    all_features = input_features + output_features
+    all_namespaces = {f.namespace for f in all_features}
+    if len(all_namespaces) > 1:
+        builder.add_warning(
+            header=f'Resolver "{resolver.fqn}" has output and input features from multiple namespaces.',
+            subheader=f"""All outputs and inputs of a resolver should be rooted in the same namespace.
+The resolver "{resolver.fqn}" references features in the namespaces {oxford_comma_list(sorted(all_namespaces))}.
+See https://docs.chalk.ai/docs/resolver-outputs for information about valid resolver outputs.
+""",
+        )
+
 
 _BUILT_IN_SUPPORTED_TYPES = {
     "datetime",
@@ -406,7 +463,15 @@ def _validate_feature_metadata_config(
             )
 
 
-def _validate_namespace_primary_key(namespace: str, features: List[UpsertFeatureGQL], builder: _LogBuilder):
+def _validate_namespace_primary_key(
+    namespace: str,
+    features: List[UpsertFeatureGQL],
+    builder: _LogBuilder,
+    singleton_namespaces: set[str],
+):
+    if namespace in singleton_namespaces:
+        return
+
     primary_features = list(f for f in features if f.scalarKind and f.scalarKind.primary)
 
     if len(primary_features) == 0:
@@ -423,6 +488,7 @@ def _validate_namespace_primary_key(namespace: str, features: List[UpsertFeature
 
 
 def validate_graph(request: UpsertGraphGQL) -> List[UpdateGraphError]:
+    singleton_namespaces = {c.name for c in request.featureClasses or [] if c.isSingleton}
     builder = _LogBuilder()
 
     # Validate the features
@@ -436,7 +502,12 @@ def validate_graph(request: UpsertGraphGQL) -> List[UpdateGraphError]:
         namespaces[feature.id.namespace].append(feature)
 
     for namespace, features in namespaces.items():
-        _validate_namespace_primary_key(namespace, features, builder)
+        _validate_namespace_primary_key(
+            namespace=namespace,
+            features=features,
+            builder=builder,
+            singleton_namespaces=singleton_namespaces,
+        )
 
     for feature in request.features or []:
         _validate_primary_key(feature, builder)
@@ -447,11 +518,21 @@ def validate_graph(request: UpsertGraphGQL) -> List[UpdateGraphError]:
 
     # Validate the resolvers
     fqn_to_feature = {f.id.fqn: f for f in request.features or []}
+
     _validate_resolver_names_unique(request.resolvers or [], builder)
     for resolver in request.resolvers or []:
         _validate_offline_resolver_has_one(resolver, builder)
-        _validate_resolver_input(resolver, builder)
+        _validate_resolver_input(
+            singleton_namespaces=singleton_namespaces,
+            resolver=resolver,
+            builder=builder,
+        )
         _validate_resolver_output(resolver, builder)
+        _validate_resolver_feature_cycles(fqn_to_feature=fqn_to_feature, resolver=resolver, builder=builder)
+        # TODO Some customers currently still do stuff like:
+        # >>> def some_resolver(uid: User.id) -> DataFrame[Transaction]: ....
+        # So don't even warn about it
+        # _validate_resolver_input_and_output_namespace(resolver, builder)
         _validate_no_feature_times_as_input(
             fqn_to_feature=fqn_to_feature,
             resolver=resolver,
