@@ -1,14 +1,19 @@
+import collections
 import dataclasses
 import json
-from typing import Any, List, Type, TypeVar, overload
+from typing import Any, List, Optional, Type, TypeVar, Union, overload
 
 import pydantic
 from pydantic.main import BaseModel
 
-from chalk.features import DataFrame, Feature, Filter, serialize_dtype
+from chalk.features import DataFrame, Feature, FeatureWrapper, Filter, serialize_dtype, unwrap_feature
 from chalk.features.feature_field import HasOnePathObj
+from chalk.features.feature_set import is_feature_set_class
+from chalk.features.pseudofeatures import PSEUDONAMESPACE
 from chalk.features.resolver import Cron, OfflineResolver, OnlineResolver, SinkResolver, StreamResolver
 from chalk.parsed.duplicate_input_gql import (
+    FeatureClassGQL,
+    UpsertCDCSourceGQL,
     UpsertDataFrameGQL,
     UpsertFeatureGQL,
     UpsertFeatureIdGQL,
@@ -29,6 +34,7 @@ from chalk.parsed.duplicate_input_gql import (
     UpsertStreamResolverParamMessageGQL,
     VersionInfoGQL,
 )
+from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
 from chalk.streams import Windowed
 from chalk.streams.types import (
     StreamResolverParam,
@@ -69,6 +75,8 @@ def _convert_type(underlying_type: Type) -> _ConvertedType:
         else:
             kind = str(underlying_type._kind).replace("typing.", "")
             return _ConvertedType(name=f"Windowed[{kind}]", bases=["CHALK_WINDOWED"])
+    if isinstance(underlying_type, FeatureWrapper):
+        return _convert_type(unwrap_feature(underlying_type).typ.underlying)
     assert isinstance(underlying_type, type), f"Underlying type is not a type: '{str(underlying_type)}'"
     base_classes = [x.__name__ for x in type.mro(underlying_type)]
     if "Enum" in base_classes:
@@ -93,72 +101,89 @@ def _convert_type(underlying_type: Type) -> _ConvertedType:
     )
 
 
-def _get_feature_id(s: Feature):
+def _get_feature_id(s: Feature, root_fqn: bool):
     assert s.features_cls is not None
-    assert s.name is not None, "Could not find feature name"
-    assert s.namespace is not None, "Could not find feature namespace"
     return UpsertFeatureIdGQL(
-        fqn=s.fqn,
+        fqn=s.root_fqn if root_fqn else s.fqn,
         name=s.name,
         namespace=s.namespace,
+        isPrimary=s.primary,
         className=s.features_cls.__name__,
         attributeName=s.attribute_name,
         explicitNamespace=s.namespace != to_snake_case(s.features_cls.__name__),
     )
 
 
+def _get_pseudofeature_id(s: Feature):
+    return UpsertFeatureIdGQL(
+        fqn=s.fqn,
+        name=s.name,
+        namespace=s.namespace,
+        isPrimary=s.primary,
+    )
+
+
 def _convert_df(df: Type[DataFrame]) -> UpsertDataFrameGQL:
     return UpsertDataFrameGQL(
-        columns=[_get_feature_id(f) for f in df.columns if not f.is_autogenerated],
-        filters=[convert_type_to_gql(f) for f in df.filters],
+        columns=[
+            _get_feature_id(f, root_fqn=True)
+            for f in df.columns
+            if not f.is_autogenerated and not f.namespace == PSEUDONAMESPACE
+        ],
+        # filters=[convert_type_to_gql(f) for f in df.filters],
+        filters=None,
     )
 
 
 def _get_path_component(pc: HasOnePathObj) -> UpsertReferencePathComponentGQL:
     assert isinstance(pc.parent, Feature), f"Parent in relationship path not a feature, but {type(pc).__name__}"
     return UpsertReferencePathComponentGQL(
-        parent=_get_feature_id(pc.parent),
-        child=_get_feature_id(pc.child),
+        parent=_get_feature_id(pc.parent, root_fqn=False),
+        child=_get_feature_id(pc.child, root_fqn=False),
         parentToChildAttributeName=pc.parent_to_child_attribute_name,
     )
 
 
 @overload
-def convert_type_to_gql(t: Filter) -> UpsertFilterGQL:
+def convert_type_to_gql(t: Filter, path_prefix: Optional[str] = None) -> UpsertFilterGQL:
     ...
 
 
 @overload
-def convert_type_to_gql(t: StreamResolver) -> UpsertStreamResolverGQL:
+def convert_type_to_gql(t: StreamResolver, path_prefix: Optional[str] = None) -> UpsertStreamResolverGQL:
     ...
 
 
 @overload
-def convert_type_to_gql(t: SinkResolver) -> UpsertSinkResolverGQL:
+def convert_type_to_gql(t: SinkResolver, path_prefix: Optional[str] = None) -> UpsertSinkResolverGQL:
     ...
 
 
 @overload
-def convert_type_to_gql(t: OnlineResolver) -> UpsertResolverGQL:
+def convert_type_to_gql(
+    t: Union[OnlineResolver, OfflineResolver], path_prefix: Optional[str] = None
+) -> UpsertResolverGQL:
     ...
 
 
 @overload
-def convert_type_to_gql(t: OfflineResolver) -> UpsertResolverGQL:
+def convert_type_to_gql(t: StreamResolverParam, path_prefix: Optional[str] = None) -> UpsertStreamResolverParamGQL:
     ...
 
 
 @overload
-def convert_type_to_gql(t: StreamResolverParam) -> UpsertStreamResolverParamGQL:
+def convert_type_to_gql(t: Feature, path_prefix: Optional[str] = None) -> UpsertFeatureGQL:
     ...
 
 
-@overload
-def convert_type_to_gql(t: Feature) -> UpsertFeatureGQL:
-    ...
-
-
-def convert_type_to_gql(t: Any):
+def convert_type_to_gql(
+    t: Any,
+    # Resolvers have their filenames attached to them. But for UI presentation, we don't
+    # want to show the full path, e.g. /Users/customer-name/projects/chalk/....,
+    # but we want to show just the path relative to the project root. To do this efficiently,
+    # we'll pass the project root path here and delete it from those full paths
+    path_prefix: Optional[str] = None,
+):
     if isinstance(t, StreamResolver):
         return UpsertStreamResolverGQL(
             fqn=t.fqn,
@@ -170,7 +195,7 @@ def convert_type_to_gql(t: Any):
             doc=t.fn.__doc__,
             machineType=t.machine_type,
             output=[
-                _get_feature_id(f)
+                _get_feature_id(f, root_fqn=True)
                 for f in t.output_features
                 if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
             ],
@@ -180,6 +205,8 @@ def convert_type_to_gql(t: Any):
                 if not isinstance(i, StreamResolverParamMessageWindow)
             ],
             owner=t.owner,
+            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
+            sourceLine=t.source_line,
         )
 
     if isinstance(t, SinkResolver):
@@ -188,7 +215,7 @@ def convert_type_to_gql(t: Any):
             functionDefinition=t.function_definition,
             inputs=[
                 UpsertFeatureReferenceGQL(
-                    underlying=_get_feature_id(f),
+                    underlying=_get_feature_id(f, root_fqn=False),
                     path=[_get_path_component(p) for p in f.path or []],
                 )
                 for f in t.inputs
@@ -203,6 +230,8 @@ def convert_type_to_gql(t: Any):
             maxDelay=None if t.max_delay is None else timedelta_to_duration(t.max_delay),
             upsert=t.upsert,
             owner=t.owner,
+            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
+            sourceLine=t.source_line,
         )
 
     if isinstance(t, (OnlineResolver, OfflineResolver)):
@@ -214,12 +243,19 @@ def convert_type_to_gql(t: Any):
         all_inputs = []
         for f in t.inputs:
             if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated:
-                feature_ref = UpsertFeatureReferenceGQL(
-                    underlying=_get_feature_id(f),
-                    path=[_get_path_component(p) for p in f.path or []],
-                )
-                all_inputs.append(UpsertResolverInputUnionGQL(feature=feature_ref))
-                inputs.append(feature_ref)
+                if f.namespace == PSEUDONAMESPACE:
+                    feature_ref = UpsertFeatureReferenceGQL(
+                        underlying=_get_pseudofeature_id(f),
+                        path=[_get_path_component(p) for p in f.path or []],
+                    )
+                    all_inputs.append(UpsertResolverInputUnionGQL(pseudoFeature=feature_ref))
+                else:
+                    feature_ref = UpsertFeatureReferenceGQL(
+                        underlying=_get_feature_id(f, root_fqn=False),
+                        path=[_get_path_component(p) for p in f.path or []],
+                    )
+                    inputs.append(feature_ref)
+                    all_inputs.append(UpsertResolverInputUnionGQL(feature=feature_ref))
             elif isinstance(f, type) and issubclass(f, DataFrame):
                 all_inputs.append(UpsertResolverInputUnionGQL(dataframe=_convert_df(f)))
 
@@ -229,10 +265,10 @@ def convert_type_to_gql(t: Any):
             kind="offline" if isinstance(t, OfflineResolver) else "online",
             functionDefinition=t.function_definition,
             inputs=inputs,
-            all_inputs=all_inputs,
+            allInputs=all_inputs,
             output=UpsertResolverOutputGQL(
                 features=[
-                    _get_feature_id(f)
+                    _get_feature_id(f, root_fqn=True)
                     for f in t.output.features
                     if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
                 ],
@@ -247,6 +283,8 @@ def convert_type_to_gql(t: Any):
             machineType=t.machine_type,
             owner=t.owner,
             timeout=None if t.timeout is None else timedelta_to_duration(t.timeout),
+            filename=t.filename if path_prefix is None else t.filename.replace(path_prefix, ""),
+            sourceLine=t.source_line,
         )
 
     if isinstance(t, StreamResolverParam):
@@ -310,6 +348,12 @@ def convert_type_to_gql(t: Any):
             feature_time_kind_gql = UpsertFeatureTimeKindGQL()
         else:
             assert t.typ is not None
+            parsed_type = t.typ.parsed_annotation
+            if hasattr(parsed_type, "__chalk_feature_set__") and parsed_type.__chalk_feature_set__:
+                raise TypeError(
+                    f"Feature '{t.fqn}' is annotated with feature set '{parsed_type}' but no join was found. "
+                    f"Please ensure that join between keys is specified between '{t.namespace}' and '{parsed_type.namespace}'."
+                )
             converted_type = _convert_type(t.typ.underlying)
             scalar_kind_gql = UpsertScalarKindGQL(
                 scalarKind=converted_type.name,
@@ -330,6 +374,7 @@ def convert_type_to_gql(t: Any):
                 fqn=t.fqn,
                 name=t.name,
                 namespace=t.namespace,
+                isPrimary=t.primary,
                 className=t.features_cls.__name__,
                 attributeName=t.attribute_name,
                 explicitNamespace=t.namespace != to_snake_case(t.features_cls.__name__),
@@ -344,7 +389,17 @@ def convert_type_to_gql(t: Any):
             hasManyKind=has_many_kind_gql,
             scalarKind=scalar_kind_gql,
             featureTimeKind=feature_time_kind_gql,
-            namespacePath=str(paths.get_classpath(t.features_cls)),
+            namespacePath=str(paths.get_classpath_or_name(t.features_cls)),
+            isSingleton=t.is_singleton,
+        )
+
+    if is_feature_set_class(t):
+        return FeatureClassGQL(
+            isSingleton=t.__chalk_is_singleton__,
+            doc=t.__doc__,
+            name=t.__chalk_namespace__,
+            owner=t.__chalk_owner__,
+            tags=t.__chalk_tags__,
         )
 
     if isinstance(t, Filter):
@@ -359,9 +414,26 @@ def convert_type_to_gql(t: Any):
             ), f"Filters must be on scalar features, but {t.rhs.fqn} is not a scalar."
 
         return UpsertFilterGQL(
-            lhs=_get_feature_id(t.lhs),
+            lhs=_get_feature_id(t.lhs, root_fqn=True),
             op=t.operation,
-            rhs=_get_feature_id(t.rhs),
+            rhs=_get_feature_id(t.rhs, root_fqn=True),
         )
 
     raise ValueError(f"Unable to convert {t} to GQL")
+
+
+def gather_cdc_sources() -> List[UpsertCDCSourceGQL]:
+    integration_name_to_tables = collections.defaultdict(list)
+    for source in BaseSQLSource.registry:
+        if isinstance(source, TableIngestMixIn):
+            for schema_dot_table, preferences in source.ingested_tables.items():
+                if preferences.cdc is True:
+                    assert isinstance(source, BaseSQLSource)
+                    integration_name_to_tables[source.name].append(schema_dot_table)
+    return [
+        UpsertCDCSourceGQL(
+            integrationName=integration_name,
+            schemaDotTableList=tables,
+        )
+        for integration_name, tables in integration_name_to_tables.items()
+    ]
