@@ -1,4 +1,4 @@
-from typing import Any, List, Mapping, Optional, Union
+from typing import Any, List, Literal, Mapping, Optional, Union
 
 from chalk.features import Feature
 from chalk.sql.finalized_query import FinalizedChalkQuery, Finalizer
@@ -22,7 +22,7 @@ class ChalkQuery(ChalkQueryProtocol):
         self._features = features
         self._source = source
 
-    def first(self, params: Optional[Mapping[str, Any]] = None):
+    def first(self, params: Optional[Mapping[str, Any]] = None) -> FinalizedChalkQuery:
         from sqlalchemy.sql import Select
 
         self._query = self._query.limit(1)
@@ -36,7 +36,7 @@ class ChalkQuery(ChalkQueryProtocol):
             fields=self._features,
         )
 
-    def one_or_none(self, params: Optional[Mapping[str, Any]] = None):
+    def one_or_none(self, params: Optional[Mapping[str, Any]] = None) -> FinalizedChalkQuery:
         from sqlalchemy.sql import Select
 
         self._query = self._query.limit(1)
@@ -50,7 +50,7 @@ class ChalkQuery(ChalkQueryProtocol):
             fields=self._features,
         )
 
-    def one(self, params: Optional[Mapping[str, Any]] = None):
+    def one(self, params: Optional[Mapping[str, Any]] = None) -> FinalizedChalkQuery:
         from sqlalchemy.sql import Select
 
         self._query = self._query.limit(1)
@@ -64,7 +64,7 @@ class ChalkQuery(ChalkQueryProtocol):
             fields=self._features,
         )
 
-    def all(self, params: Optional[Mapping[str, Any]] = None):
+    def all(self, params: Optional[Mapping[str, Any]] = None) -> FinalizedChalkQuery:
         from sqlalchemy.sql import Select
 
         assert isinstance(self._query.selectable, Select), "A text query's selectable should always be a Select"
@@ -80,10 +80,11 @@ class ChalkQuery(ChalkQueryProtocol):
     def incremental(
         self,
         lookback_period: Duration = "0s",
-        mode: str = "row",
+        mode: Literal["row", "group", "parameter"] = "row",
         incremental_column: Optional[Union[str, Feature]] = None,
+        incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
         params: Optional[Mapping[str, Any]] = None,
-    ):
+    ) -> FinalizedChalkQuery:
         from sqlalchemy.sql import Select
 
         assert isinstance(self._query.selectable, Select)
@@ -97,6 +98,7 @@ class ChalkQuery(ChalkQueryProtocol):
                 ),
                 incremental_column=None if incremental_column is None else str(incremental_column),
                 mode=mode,
+                incremental_timestamp=incremental_timestamp,
             ),
             params=params or {},
             fields=self._features,
