@@ -4,7 +4,7 @@ import os
 from typing import TYPE_CHECKING, Any, Dict, Optional, Union
 
 from chalk.integrations.named import load_integration_variable
-from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
+from chalk.sql._internal.sql_source import BaseSQLSource, SQLSourceKind, TableIngestMixIn
 from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
 from chalk.utils.missing_dependency import missing_dependency_exception
 
@@ -22,6 +22,7 @@ class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestP
         password: Optional[str] = None,
         name: Optional[str] = None,
         engine_args: Optional[Dict[str, Any]] = None,
+        async_engine_args: Optional[Dict[str, Any]] = None,
     ):
         try:
             import pymysql
@@ -41,15 +42,25 @@ class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestP
         self.ingested_tables: Dict[str, Any] = {}
         if engine_args is None:
             engine_args = {}
+        if async_engine_args is None:
+            async_engine_args = {}
         engine_args.setdefault("pool_size", 20)
         engine_args.setdefault("max_overflow", 60)
+        async_engine_args.setdefault("pool_size", 20)
+        async_engine_args.setdefault("max_overflow", 60)
         # We set the default isolation level to autocommit since the SQL sources are read-only, and thus
         # transactions are not needed
         # Setting the isolation level on the engine, instead of the connection, avoids
         # a DBAPI statement to reset the transactional level back to the default before returning the
         # connection to the pool
         engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
-        BaseSQLSource.__init__(self, name=name, engine_args=engine_args)
+        async_engine_args.setdefault("isolation_level", os.environ.get("CHALK_SQL_ISOLATION_LEVEL", "AUTOCOMMIT"))
+        BaseSQLSource.__init__(self, name=name, engine_args=engine_args, async_engine_args=async_engine_args)
+
+    kind = SQLSourceKind.mysql
+
+    def get_sqlglot_dialect(self) -> str | None:
+        return "mysql"
 
     def local_engine_url(self) -> URL:
         from sqlalchemy.engine.url import URL
@@ -62,3 +73,15 @@ class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestP
             port=self.port,
             database=self.db,
         )
+
+    def async_local_engine_url(self) -> URL:
+        from sqlalchemy.engine.url import URL
+
+        return URL.create(
+            drivername="mysql+aiomysql",
+            username=self.user,
+            password=self.password,
+            host=self.host,
+            port=self.port,
+            database=self.db,
+        )
