@@ -6,7 +6,7 @@ import json
 import os
 import re
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Literal, Optional, Type, Union, cast
+from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Literal, Mapping, Optional, Sequence, Type, Union, cast
 
 import yaml
 from pydantic import BaseModel, ValidationError, validator
@@ -27,12 +27,12 @@ from chalk.streams.base import StreamSource
 from chalk.streams.types import StreamResolverSignature
 from chalk.utils.duration import parse_chalk_duration
 from chalk.utils.missing_dependency import missing_dependency_exception
-from chalk.utils.string import to_camel_case
+from chalk.utils.string import to_snake_case
 
 if TYPE_CHECKING:
     from sqlglot.expressions import Expression
 
-_SOURCES = {
+_SOURCES: Mapping[str, Union[Type[BaseSQLSource], Type[StreamSource]]] = {
     "snowflake": SnowflakeSourceImpl,
     "postgres": PostgreSQLSourceImpl,
     "postgresql": PostgreSQLSourceImpl,
@@ -44,14 +44,19 @@ _SOURCES = {
     "kafka": KafkaSource,
 }
 
-_DIALECTS_SQLGLOT: Dict[Type[BaseSQLSource], str] = {
-    SnowflakeSourceImpl: "snowflake",
-    PostgreSQLSourceImpl: "postgres",
-    MySQLSourceImpl: "mysql",
-    BigQuerySourceImpl: "bigquery",
-    RedshiftSourceImpl: "redshift",
-    SQLiteSourceImpl: "sqlite",
-}
+
+_SQLGLOT_DIALECTS = frozenset(
+    (
+        "snowflake",
+        "postgres",
+        "mysql",
+        "bigquery",
+        "redshift",
+        "sqlite",
+        "databricks",
+    )
+)
+"""These dialects are used if a "kind" of source is listed, rather than a specific source."""
 
 _RESOLVER_TYPES = {
     "offline": OfflineResolver,
@@ -67,6 +72,7 @@ class IncrementalSettings(BaseModel):
     incremental_column: Optional[str]
     lookback_period: Optional[str]
     mode: Literal["row", "group", "parameter"]
+    incremental_timestamp: Optional[Literal["feature_time", "resolver_execution_time"]] = "feature_time"
 
     @validator("lookback_period")
     @classmethod
@@ -175,7 +181,7 @@ class GlotResult:
     sql_string: str
     glot: Optional[Expression]
     args: Dict[str, str]
-    default_args: List[Optional[str]]
+    default_args: List[Union[Optional[str], ellipsis]]
     comment_dict: Optional[CommentDict]
     docstring: Optional[str]
     errors: List[ResolverError]
@@ -195,7 +201,8 @@ class ParseResult:
 
 
 def get_sql_file_resolvers(
-    sql_file_resolve_location: Path, sources: Iterable[BaseSQLSource]
+    sql_file_resolve_location: Path,
+    sources: Sequence[BaseSQLSource],
 ) -> Iterable[ResolverResult]:
     """Iterate through all `.chalk.sql` filepaths, gather the sql strings, and get a resolver hopefully for each."""
     for dp, dn, fn in os.walk(os.path.expanduser(sql_file_resolve_location)):
@@ -208,28 +215,63 @@ def get_sql_file_resolvers(
             yield get_sql_file_resolver(sources, sql_string_result)
 
 
+def get_sql_file_resolvers_from_paths(
+    sources: Sequence[BaseSQLSource],
+    paths: List[str],
+) -> List[ResolverResult]:
+    return [
+        get_sql_file_resolver(
+            sources=sources,
+            sql_string_result=_get_sql_string(p),
+        )
+        for p in paths
+    ]
+
+
 def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: SQLStringResult) -> ResolverResult:
     """Parse the sql strings and get a ResolverResult from each"""
     if sql_string_result.error:
-        return ResolverResult(resolver=None, errors=[sql_string_result.error], db=None, fields=None, args=None)
+        return ResolverResult(
+            resolver=None,
+            errors=[sql_string_result.error],
+            db=None,
+            fields=None,
+            args=None,
+        )
     assert sql_string_result.sql_string is not None
     path = sql_string_result.path
 
     errors: List[ResolverError] = []
-    glot_result = _get_sql_glot(sql_string=sql_string_result.sql_string, path=path, sources=sources)
+    glot_result = _get_sql_glot(
+        sql_string=sql_string_result.sql_string,
+        path=path,
+        sources=sources,
+    )
     if glot_result.errors:
-        return ResolverResult(resolver=None, errors=glot_result.errors, db=None, fields=None, args=None)
+        return ResolverResult(
+            resolver=None,
+            errors=glot_result.errors,
+            db=None,
+            fields=None,
+            args=None,
+        )
 
     parsed = _parse_glot(glot_result, path, sources)
     if parsed.errors:
-        return ResolverResult(resolver=None, errors=parsed.errors, db=parsed.source, fields=None, args=None)
+        return ResolverResult(
+            resolver=None,
+            errors=parsed.errors,
+            db=parsed.source,
+            fields=None,
+            args=None,
+        )
 
     # validate inputs and outputs as real features in graph
     inputs: List[Feature] = []
     for arg in glot_result.args.values():
         try:
             inputs.append(Feature.from_root_fqn(arg))
-        except Exception as e:
+        except Exception:
             errors.append(
                 ResolverError(
                     display=f"The file '{path}' references an input feature '{arg}' which does not exist",
@@ -238,16 +280,19 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
                 )
             )
     outputs: List[Feature] = []
-    for output in parsed.fields.values():
+    query_fields: Dict[str, str] = {}
+    for variable, output in parsed.fields.items():
         if output.endswith("*"):
             features = FeatureSetBase.registry[parsed.namespace].features
             for feature in features:
                 if feature.is_scalar:
                     outputs.append(feature)
             continue
+        query_fields[variable] = output
+
         try:
             outputs.append(Feature.from_root_fqn(output))
-        except Exception as e:
+        except Exception:
             errors.append(
                 ResolverError(
                     display=f"The file '{path}' references an output feature '{output}' which does not exist",
@@ -272,7 +317,7 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
         *input_values: Any,
         database: BaseSQLSource = source,
         sql_query: str = parsed.sql_string,
-        field_dict: Dict[str, str] = parsed.fields,
+        field_dict: Dict[str, str] = query_fields,
         args_dict: Dict[str, str] = glot_result.args,
         incremental: Optional[Dict[str, Any]] = incremental_dict,
     ):
@@ -301,7 +346,8 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
         output = Features[DataFrame[tuple(outputs)]]
 
     default_args: List[Optional[ResolverArgErrorHandler]] = [
-        ResolverArgErrorHandler(default_value) for default_value in glot_result.default_args
+        None if default_value is ... else ResolverArgErrorHandler(default_value)
+        for default_value in glot_result.default_args
     ]
 
     filename = os.path.basename(path)
@@ -310,7 +356,7 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
         assert resolver_type in (OnlineResolver, OfflineResolver)
         resolver = resolver_type(
             filename=path,
-            function_definition=_remove_comments(sql_string_result.sql_string),
+            function_definition=sql_string_result.sql_string,
             fqn=filename.replace(".chalk.sql", ""),
             doc=parsed.docstring,
             inputs=inputs,
@@ -326,6 +372,7 @@ def get_sql_file_resolver(sources: Iterable[BaseSQLSource], sql_string_result: S
             default_args=default_args,
             owner=parsed.comment_dict.owner,
             timeout=parsed.comment_dict.timeout,
+            is_sql_file_resolver=True,
         )
     except Exception as e:
         errors.append(
@@ -377,7 +424,7 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
     args = {}  # sql string -> input feature string
     variables = set(re.findall("\\${.*?\\}", sql_string))
     errors: List[ResolverError] = []
-    default_args: List[Optional[str]] = []
+    default_args: List[Union[Optional[str], ellipsis]] = []
     # replace ?{variable_name} with :variable_name for sqlalchemy, and keep track of input args necessary
     for variable_pattern in variables:
         has_default_arg = False
@@ -406,8 +453,7 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
                     default_args.append(default_arg)
                     has_default_arg = True
         if not has_default_arg:
-            # TODO default value of None should be refactored since None is a legitimate default value
-            default_args.append(None)
+            default_args.append(...)
         period_replaced = variable.replace(".", "_")
         sql_safe_str = f"__chalk_{period_replaced}__"
         sql_string = sql_string.replace(variable_pattern, f":{sql_safe_str}")
@@ -428,6 +474,24 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
         else:
             break
 
+    if len(comments) == 0:
+        errors.append(
+            ResolverError(
+                display="SQL file resolvers require comments that describe key-value pairs in YAML form. "
+                "See https://docs.chalk.ai/docs/sql#sql-file-resolvers for reference",
+                path=path,
+                parameter=None,
+            )
+        )
+        return GlotResult(
+            sql_string=sql_string,
+            glot=None,
+            args=args,
+            default_args=default_args,
+            comment_dict=None,
+            docstring=docstring,
+            errors=errors,
+        )
     try:
         comment_dict: Dict[str, Any] = yaml.safe_load(comments)
     except Exception as e:
@@ -453,7 +517,7 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
         for error in e.errors():
             errors.append(
                 ResolverError(
-                    display=f"Could not parse comment(s) '{'.'.join(str(x) for x in error['loc'])}': {error['msg']}",
+                    display=f"Could not parse comment(s) for SQL file resolver at '{path}': '{'.'.join(str(x) for x in error['loc'])}': {error['msg']}",
                     path=path,
                     parameter=json.dumps(comment_dict),
                 )
@@ -473,10 +537,14 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
     source_type = None
     if source_comment not in _SOURCES:  # actual name of source
         for possible_source in sources:
-            if possible_source.name == source_comment:
-                source_type = _DIALECTS_SQLGLOT.get(type(possible_source))  # dialect may be unsupported
+            if possible_source.name == source_comment and isinstance(possible_source, BaseSQLSource):
+                # Note: it is apparently possible for non-BaseSQLSources to be passed here, including
+                # e.g. `KafkaSource`.
+                source_type = (
+                    possible_source.get_sqlglot_dialect()
+                )  # dialect may be unsupported, in which case this returns `None`
     else:
-        if source_comment in _DIALECTS_SQLGLOT.values():  # source type, if supported
+        if source_comment in _SQLGLOT_DIALECTS:  # source type, if supported
             source_type = source_comment
     try:
         glot = sqlglot.parse_one(sql=sql_string, read=source_type)
@@ -508,6 +576,11 @@ def _get_sql_glot(sql_string: str, path: str, sources: Iterable[BaseSQLSource])
 
 def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSource]) -> ParseResult:
     """Parse useful info from sqlglot and gracefully exit if unable to"""
+    try:
+        import sqlglot
+        import sqlglot.expressions
+    except ImportError:
+        raise missing_dependency_exception("chalkpy[runtime]")
     # define a source SQL database. Can either specify name or kind if only one of the kind is present.
     comment_dict = glot_result.comment_dict
     assert comment_dict is not None
@@ -539,16 +612,34 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
     # get resolver fields: which columns selected will match to which chalk feature?
     namespace = comment_dict.resolves
     if namespace not in FeatureSetBase.registry:
-        camelcase_namespace = to_camel_case(namespace)
-        if camelcase_namespace in FeatureSetBase.registry:
-            namespace = camelcase_namespace
+        snaked_namespace = to_snake_case(namespace)
+        if snaked_namespace in FeatureSetBase.registry:
+            namespace = snaked_namespace
         else:
             errors.append(
                 ResolverError(
                     display=f"No @features class with the name '{namespace}'", path=path, parameter=source_name
                 )
             )
-    if errors:
+
+    if glot_result.glot is None:
+        errors.append(
+            ResolverError(
+                display=f"Could not parse SQL query {glot_result.sql_string}.",
+                path=path,
+                parameter=source_name,
+            )
+        )
+    elif not isinstance(glot_result.glot, sqlglot.expressions.Select):
+        errors.append(
+            ResolverError(
+                display=f"SQL query {glot_result.sql_string} should be a 'SELECT'",
+                path=path,
+                parameter=None,
+            )
+        )
+
+    if len(errors) > 0:
         return ParseResult(
             sql_string=glot_result.sql_string,
             comment_dict=comment_dict,
@@ -559,10 +650,11 @@ def _parse_glot(glot_result: GlotResult, path: str, sources: Iterable[BaseSQLSou
             errors=errors,
         )
 
-    fields: Dict[str, str] = {}  # sql string -> output feature string
-    assert glot_result.glot is not None
-    for column_name in glot_result.glot.named_selects:
-        fields[column_name] = f"{namespace}.{column_name}"
+    assert isinstance(glot_result.glot, sqlglot.expressions.Select)
+    # sql target -> output feature string
+    fields: Dict[str, str] = {
+        column_name: f"{namespace}.{column_name}" for column_name in glot_result.glot.named_selects
+    }
 
     return ParseResult(
         sql_string=glot_result.sql_string,
@@ -619,6 +711,8 @@ def _get_stream_resolver(
             sql_query=sql_query,
             owner=parsed.comment_dict.owner,
             parse=None,
+            keys=None,  # TODO implement parse and keys for sql file resolvers?
+            timestamp=None,
         )
     except Exception as e:
         errors.append(
