@@ -27,10 +27,15 @@ from chalk.utils.duration import Duration
 
 if TYPE_CHECKING:
     import sqlalchemy
-    from sqlalchemy.engine import URL, Connection
+    import sqlalchemy.ext.asyncio
+    from sqlalchemy.engine import Connection
     from sqlalchemy.orm import Session
 
-    from chalk.sql.finalized_query import FinalizedChalkQuery
+    from chalk.sql.finalized_query import (
+        DataframeFinalizedChalkQuery,
+        FinalizedChalkQuery,
+        SingletonFinalizedChalkQuery,
+    )
 
 TTableIngestProtocol = TypeVar("TTableIngestProtocol", bound="TableIngestProtocol")
 
@@ -54,7 +59,7 @@ class StringChalkQueryProtocol(Protocol):
         assert isinstance(ans, DataFrame)
         return ans
 
-    def one_or_none(self) -> FinalizedChalkQuery:
+    def one_or_none(self) -> SingletonFinalizedChalkQuery:
         """Return at most one result or raise an exception.
 
         Returns `None` if the query selects no rows. Raises if
@@ -64,27 +69,27 @@ class StringChalkQueryProtocol(Protocol):
 
         Returns
         -------
-        FinalizedChalkQuery
+        SingletonFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
-    def one(self) -> FinalizedChalkQuery:
+    def one(self) -> SingletonFinalizedChalkQuery:
         """Return exactly one result or raise an exception.
 
         Returns
         -------
-        FinalizedChalkQuery
+        SingletonFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
-    def all(self) -> FinalizedChalkQuery:
+    def all(self) -> DataframeFinalizedChalkQuery:
         """Return the results represented by this Query as a list.
 
         Returns
         -------
-        FinalizedChalkQuery
+        DataframeFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
@@ -96,7 +101,8 @@ class StringChalkQueryProtocol(Protocol):
         incremental_column: str,
         lookback_period: Duration = "0s",
         mode: Literal["row", "group"] = "row",
-    ) -> FinalizedChalkQuery:
+        incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
+    ) -> DataframeFinalizedChalkQuery:
         ...
 
     @overload
@@ -106,7 +112,8 @@ class StringChalkQueryProtocol(Protocol):
         incremental_column: None = None,
         lookback_period: None = None,
         mode: Literal["parameter"],
-    ) -> FinalizedChalkQuery:
+        incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
+    ) -> DataframeFinalizedChalkQuery:
         ...
 
     def incremental(
@@ -115,7 +122,8 @@ class StringChalkQueryProtocol(Protocol):
         incremental_column: Optional[str] = None,
         lookback_period: Optional[Duration] = "0s",
         mode: Literal["row", "group", "parameter"] = "row",
-    ) -> FinalizedChalkQuery:
+        incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
+    ) -> DataframeFinalizedChalkQuery:
         """Operates like `.all()`, but tracks `previous_latest_row_timestamp` between query executions in
         order to limit the amount of data returned.
 
@@ -180,27 +188,30 @@ class StringChalkQueryProtocol(Protocol):
             Defaults to `"row"`, which indicates that only rows newer than the last observed row should be
             considered. When set to `"group"`, Chalk will only ingest features from groups which are newer
             than the last observation time. This requires that the query is grouped by a primary key.
-
+        incremental_timestamp
+            Defaults to `"feature_time"`, which means that the timestamp associated with the last feature value
+            will be used as the incremental time. Alternatively, setting this parameter to `"resolver_execution_time"`
+            will use last literal timestamp that the resolver ran.
         Returns
         -------
-        FinalizedChalkQuery
+        DataframeFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
 
 class ChalkQueryProtocol(Protocol):
-    def first(self) -> FinalizedChalkQuery:
+    def first(self) -> SingletonFinalizedChalkQuery:
         """Return the first result of this Query or None if the result doesn't contain any row.
 
         Returns
         -------
-        FinalizedChalkQuery
+        SingletonFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
-    def one_or_none(self) -> FinalizedChalkQuery:
+    def one_or_none(self) -> SingletonFinalizedChalkQuery:
         """Return at most one result or raise an exception.
 
         Returns `None` if the query selects no rows. Raises if
@@ -210,27 +221,27 @@ class ChalkQueryProtocol(Protocol):
 
         Returns
         -------
-        FinalizedChalkQuery
+        SingletonFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
-    def one(self) -> FinalizedChalkQuery:
+    def one(self) -> SingletonFinalizedChalkQuery:
         """Return exactly one result or raise an exception.
 
         Returns
         -------
-        FinalizedChalkQuery
+        SingletonFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
 
-    def all(self) -> FinalizedChalkQuery:
+    def all(self) -> DataframeFinalizedChalkQuery:
         """Return the results represented by this Query as a `DataFrame`.
 
         Returns
         -------
-        FinalizedChalkQuery
+        DataframeFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
@@ -238,9 +249,11 @@ class ChalkQueryProtocol(Protocol):
     def incremental(
         self,
         lookback_period: Duration = "0s",
-        mode: str = "row",
+        mode: Literal["row", "group", "parameter"] = "row",
         incremental_column: Optional[Union[str, Feature]] = None,
-    ) -> FinalizedChalkQuery:
+        incremental_timestamp: Literal["feature_time", "resolver_execution_time"] = "feature_time",
+        params: Optional[Mapping[str, Any]] = None,
+    ) -> DataframeFinalizedChalkQuery:
         """Operates like `.all()`, but tracks `previous_latest_row_timestamp` between query executions in
         order to limit the amount of data returned.
 
@@ -305,10 +318,13 @@ class ChalkQueryProtocol(Protocol):
             Defaults to `"row"`, which indicates that only rows newer than the last observed row should be
             considered. When set to `"group"`, Chalk will only ingest features from groups which are newer
             than the last observation time. This requires that the query is grouped by a primary key.
-
+        incremental_timestamp
+            Defaults to `"feature_time"`, which means that the timestamp associated with the last feature value
+            will be used as the incremental time. Alternatively, setting this parameter to `"resolver_execution_time"`
+            will use last literal timestamp that the resolver ran.
         Returns
         -------
-        FinalizedChalkQuery
+        DataframeFinalizedChalkQuery
             A query that can be returned from a resolver.
         """
         ...
@@ -402,10 +418,14 @@ class ChalkQueryProtocol(Protocol):
         a resolver as an intermediate feature, and processing that
         intermediate feature in a different resolver.
 
+        Note: this requires the usage of the `fields={...}` argument when used in conjunction with `query_string`
+        or `query_sql_file`.
+
         Returns
         -------
         Any
-            The raw result of executing the query.
+            The raw result of executing the query. For `.all()`, returns a `DataFrame`. For `.one()` or `.one_or_none()`,
+            returns a `Features` instance corresponding to the relevant feature class.
         """
         return self.all().execute()
 
@@ -501,6 +521,38 @@ class BaseSQLSourceProtocol(Protocol):
         finalized_query: FinalizedChalkQuery,
         columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
         connection: Optional[Connection] = None,
+        attempt_efficient_execution: bool = True,
+        polars_read_csv: bool = False,
+    ) -> pa.Table:
+        """Execute a query to a `pa.Table`.
+
+        Parameters
+        ----------
+        finalized_query
+            The query to execute
+        columns_to_converters
+            A function that, given the list of column names returned from the database,
+            produces a mapping of columns to converters. Columns not in the resulting dictionary should be ignored.
+        connection
+            A connection to execute the query under. If not specified, then a new connection is acquired.
+        attempt_efficient_execution
+            The executor will attempt to avoid the use of SQLAlchemy and use more efficient methods to
+            retrieve data.
+
+        Returns
+        -------
+        pa.Table
+            A `pa.Table` containing the results.
+        """
+        ...
+
+    async def async_execute_query(
+        self,
+        finalized_query: FinalizedChalkQuery,
+        columns_to_converters: Callable[[List[str]], Dict[str, FeatureConverter]],
+        connection: Optional[sqlalchemy.ext.asyncio.AsyncConnection] = None,
+        attempt_efficient_execution: bool = True,
+        polars_read_csv: bool = False,
     ) -> pa.Table:
         """Execute a query to a `pa.Table`.
 
@@ -531,6 +583,15 @@ class BaseSQLSourceProtocol(Protocol):
         """
         ...
 
+    def get_sqlglot_dialect(self) -> Union[str, None]:
+        """Returns dialect for sqlglot operations.  Also happens to be the name of the source.
+
+        Returns
+        -------
+        Union[str, None]
+            The dialect if applicable to sqlglot
+        """
+
 
 class TableIngestProtocol(Protocol):
     def with_table(
