@@ -1,5 +1,5 @@
 import inspect
-from typing import Callable, Literal, Optional, TypeVar
+from typing import Any, Callable, Dict, Literal, Optional, TypeVar
 
 from typing_extensions import ParamSpec
 
@@ -7,6 +7,7 @@ from chalk.features.tag import Environments
 from chalk.state import KeyedState
 from chalk.streams._file_source import FileSource
 from chalk.streams._kafka_source import KafkaSource
+from chalk.streams._kinesis_source import KinesisSource
 from chalk.streams._windows import Windowed, get_duration_secs, get_name_with_duration, windowed
 from chalk.streams.base import StreamSource
 from chalk.utils import AnyDataclass, MachineType
@@ -15,12 +16,13 @@ __all__ = [
     "FileSource",
     "KafkaSource",
     "KeyedState",
+    "KinesisSource",
     "StreamSource",
     "Windowed",
+    "get_duration_secs",
+    "get_name_with_duration",
     "stream",
     "windowed",
-    "get_name_with_duration",
-    "get_duration_secs",
 ]
 
 P = ParamSpec("P")
@@ -34,17 +36,19 @@ def stream(
     mode: Optional[Literal["continuous", "tumbling"]] = None,
     environment: Optional[Environments] = None,
     machine_type: Optional[MachineType] = None,
-    message: Optional[AnyDataclass] = None,
     owner: Optional[str] = None,
     parse: Optional[Callable[[T], V]] = None,
+    keys: Optional[Dict[str, Any]] = None,
+    timestamp: Optional[str] = None,
 ) -> Callable[[Callable[P, T]], Callable[P, T]]:
     """Decorator to create a stream resolver.
 
     Parameters
     ----------
     source
-        The streaming source, e.g. `KafkaSource()`.
+        The streaming source, e.g. `KafkaSource(...)` or `KinesisSource(...)`.
     mode
+        This parameter is defined when the streaming resolver returns a windowed feature.
         Tumbling windows are fixed-size, contiguous and non-overlapping time intervals. You can think of
         tumbling windows as adjacently arranged bins of equal width.
         Tumbling windows are most often used alongside `max_staleness` to allow the features
@@ -55,6 +59,24 @@ def stream(
         at all the messages received in the window and computes the value on-demand.
 
         See more at https://docs.chalk.ai/docs/aggregations#window-modes
+    parse
+        A callable that will interpret an input prior to the invocation of the resolver.
+        Parse functions can serve many functions, including pre-parsing bytes,
+        skipping unrelated messages, or supporting rekeying.
+
+        See more at https://docs.chalk.ai/docs/streams#parsing
+    keys
+        A mapping from input `BaseModel` attribute to Chalk feature attribute to support continuous streaming re-keying.
+        This parameter is required for continuous resolvers.
+        Features that are included here do not have to be explicitly returned in the stream resolver:
+        the feature will automatically be set to the key value used for aggregation.
+        See more at https://docs.chalk.ai/docs/aggregations#keys
+    timestamp
+        An optional string specifying an input attribute as the timestamp used for windowed aggregations.
+        See more at https://docs.chalk.ai/docs/aggregations#custom-event-timestamping
+
+    Other Parameters
+    ----------------
     environment
         Environments are used to trigger behavior in different deployments
         such as staging, production, and local development.
@@ -65,40 +87,26 @@ def stream(
             - `list[str]` - run in any of the specified environment and no others
 
         Read more at https://docs.chalk.ai/docs/resolver-environments
-    message
-        The type of the message on a stream.
-
-        You can use any `dataclass`. The message on your stream is assumed
-        to be Json that matches the schema provided by the `dataclass`.
-        Note that Chalk does not use `dataclass` in practice to deserialize
-        the data, but instead derives a `pyarrow.Schema` that mirrors the
-        `dataclass`.
-    owner
-        Individual or team responsible for this resolver.
-        The Chalk Dashboard will display this field, and alerts
-        can be routed to owners.
-    parse
-        A callable that will interpret an input prior to the invocation of the resolver
-
-    Other Parameters
-    ----------------
     machine_type
         You can optionally specify that resolvers need to run
         on a machine other than the default. Must be configured
         in your deployment.
+    owner
+        Individual or team responsible for this resolver.
+        The Chalk Dashboard will display this field, and alerts
+        can be routed to owners.
 
     Returns
     -------
     Callable[[Any, ...], Any]
-        A callable function! You can unit-test resolvers as you would
+        A callable function! You can unit-test stream resolvers as you would
         unit-test any other code.
-
-        Read more at https://docs.chalk.ai/docs/unit-tests
     """
     caller_frame = inspect.stack()[1]
     caller_filename = caller_frame.filename
     caller_globals = caller_frame.frame.f_globals
     caller_locals = caller_frame.frame.f_locals
+    caller_line = caller_frame.frame.f_lineno
     from chalk.features.resolver import parse_and_register_stream_resolver
 
     def decorator(fn: Callable[P, T]) -> Callable[P, T]:
@@ -109,11 +117,14 @@ def stream(
             source=source,
             mode=mode,
             caller_filename=caller_filename,
+            caller_line=caller_line,
             environment=environment,
             machine_type=machine_type,
-            message=message,
+            message=None,
             owner=owner,
             parse=parse,
+            keys=keys,
+            timestamp=timestamp,
         )
 
     return decorator
