@@ -1,5 +1,6 @@
 from typing import Any, List, Literal, Optional, Union
 
+import pydantic
 from pydantic import BaseModel
 
 from chalk.integrations.named import load_integration_variable
@@ -13,7 +14,7 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
     bootstrap_server: Optional[Union[str, List[str]]] = None
     """The URL of one of your Kafka brokers from which to fetch initial metadata about your Kafka cluster"""
 
-    topic: Optional[Union[str, List[str]]] = None
+    topic: Optional[str] = None
     """The name of the topic to subscribe to."""
 
     ssl_keystore_location: Optional[str] = None
@@ -54,7 +55,7 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
     Username for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
     """
 
-    sasl_password: Optional[str] = None
+    sasl_password: Optional[str] = pydantic.Field(default=None, repr=False)
     """
     Password for SASL PLAIN, SCRAM-SHA-256, or SCRAM-SHA-512 authentication.
     """
@@ -69,11 +70,16 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
     Messages older than this deadline will not be processed.
     """
 
+    dead_letter_queue_topic: Optional[str] = None
+    """
+    Kafka topic to send messages when message processing fails
+    """
+
     def __init__(
         self,
         *,
         bootstrap_server: Optional[Union[str, List[str]]] = None,
-        topic: Optional[Union[str, List[str]]] = None,
+        topic: Optional[str] = None,
         ssl_keystore_location: Optional[str] = None,
         ssl_ca_file: Optional[str] = None,
         client_id_prefix: Optional[str] = None,
@@ -84,14 +90,14 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
         sasl_password: Optional[str] = None,
         name: Optional[str] = None,
         late_arrival_deadline: Duration = "infinity",
+        dead_letter_queue_topic: Optional[str] = None,
     ):
         super(KafkaSource, self).__init__(
             bootstrap_server=bootstrap_server
             or load_integration_variable(
                 name="KAFKA_BOOTSTRAP_SERVER", integration_name=name, parser=comma_whitespace_split
             ),
-            topic=topic
-            or load_integration_variable(name="KAFKA_TOPIC", integration_name=name, parser=comma_whitespace_split),
+            topic=topic or load_integration_variable(name="KAFKA_TOPIC", integration_name=name),
             ssl_keystore_location=ssl_keystore_location
             or load_integration_variable(name="KAFKA_SSL_KEYSTORE_LOCATION", integration_name=name),
             client_id_prefix=client_id_prefix
@@ -110,8 +116,22 @@ class KafkaSource(StreamSource, SinkIntegrationProtocol, BaseModel, frozen=True)
             sasl_password=sasl_password or load_integration_variable(name="KAFKA_SASL_PASSWORD", integration_name=name),
             name=name,
             late_arrival_deadline=late_arrival_deadline,
+            dead_letter_queue_topic=dead_letter_queue_topic,
             ssl_ca_file=ssl_ca_file or load_integration_variable(name="KAFKA_SSL_CA_FILE", integration_name=name),
         )
+        self.registry.append(self)
 
     def _config_to_json(self) -> Any:
         return self.json()
+
+    @property
+    def streaming_type(self) -> str:
+        return "kafka"
+
+    @property
+    def dlq_name(self) -> Union[str, None]:
+        return self.dead_letter_queue_topic
+
+    @property
+    def stream_or_topic_name(self) -> str:
+        return self.topic
