@@ -19,7 +19,9 @@ class ResolverAnnotationParser:
         parsed_source = ast.parse(textwrap.dedent(source))
         assert len(parsed_source.body) == 1
         function_def = parsed_source.body[0]
-        assert isinstance(function_def, ast.FunctionDef)
+        assert isinstance(
+            function_def, (ast.FunctionDef, ast.AsyncFunctionDef)
+        ), f"The resolver must be a function. Received:\n\n{source}"
         args = function_def.args
         all_args = [*args.posonlyargs, *args.args, *args.kwonlyargs]
         return all_args
