@@ -177,3 +177,15 @@ def get_unique_item_or_none(iterable: Iterable[Optional[T]]) -> Optional[T]:
             raise ValueError(f"Multiple values not permitted. Found {item}, {x}")
         item = x
     return item
+
+
+def get_unique_item_if_exists(collection: Iterable[Optional[T]]) -> Optional[T]:
+    item = None
+    for x in collection:
+        if x is None:
+            continue
+        if item is not None:
+            if x != item:
+                raise ValueError(f"Multiple values not permitted. Found {item}, {x}")
+        item = x
+    return item
