@@ -92,7 +92,7 @@ Logger = get_logger
 
 
 @contextlib.contextmanager
-def add_logging_context(_merge: bool = True, **extra: Any):
+def add_logging_context(*, _merge: bool = True, **extra: Any):
     """A context manager to push and pop `extra` dictionary keys.
 
     Parameters
