@@ -1,7 +1,7 @@
 import os
 import sys
 from pathlib import Path
-from typing import List, Optional
+from typing import Any, List, Optional
 
 
 def get_classpath(x: object) -> Path:
@@ -10,6 +10,11 @@ def get_classpath(x: object) -> Path:
     return Path(os.path.abspath(filepath))
 
 
+def get_classpath_or_name(x: Any) -> str:
+    module = sys.modules[x.__module__]
+    return getattr(module, "__file__", None) or getattr(module, "__name__")
+
+
 def get_directory_root() -> Optional[Path]:
     current = Path(os.path.dirname(os.path.abspath("dummy.txt")))
     while True:
