@@ -8,8 +8,9 @@ from typing import Dict, List, NamedTuple, Optional, Tuple, Type, Union
 
 from typing_extensions import Annotated, get_args, get_origin
 
-from chalk.features import Feature, Features, FeatureSetBase
-from chalk.importer import get_directory_root, import_all_python_files_from_dir
+from chalk.features import Feature, Features, FeatureSetBase, FeatureWrapper, unwrap_feature
+from chalk.importer import import_all_python_files_from_dir
+from chalk.utils.paths import get_directory_root
 from chalk.streams import Windowed
 
 try:
@@ -55,6 +56,8 @@ class StubGenerator:
         """
         if isinstance(item, typing._SpecialForm):
             item = str(item)
+        if isinstance(item, FeatureWrapper):
+            item = type(unwrap_feature(item))
         if isinstance(item, Windowed):
             item = type(item)
         if isinstance(item, type):
@@ -85,11 +88,12 @@ class StubGenerator:
         item = f"__stubgen_{'_'.join(item_split)}"
         return item
 
-    def _register_imports_for_annotation(self, annotation: Union[type, Windowed]) -> str:
+    def _register_imports_for_annotation(self, annotation: Union[type, Windowed]) -> Tuple[str, str]:
+        """Register all necessary imports, and return a tuple of (protocol annotation, init annotation)"""
         if annotation == type(None):
-            return "None"
+            return "None", "None"
         if annotation is ...:
-            return "..."
+            return "...", "..."
         origin, args = get_origin(annotation), get_args(annotation)
         if origin is None:
             assert args == ()
@@ -97,23 +101,33 @@ class StubGenerator:
             if isinstance(annotation, Windowed):
                 origin_name = self._add_import(annotation, import_as="Windowed")
                 assert annotation._kind is not None
-                arg_name = self._register_imports_for_annotation(annotation._kind)
-                return f"{origin_name}[{arg_name}]"
+                protocol_arg_name, init_arg_name = self._register_imports_for_annotation(annotation._kind)
+                # FIXME: I don't think this is entirely correct, b/c if a user manually constructs a windowed
+                # feature class, then it's unclear what should be passed in for the windowed arg in the init
+                # parameter
+                return f"{origin_name}[{protocol_arg_name}]", f"{origin_name}[{init_arg_name}]"
             elif annotation.__module__ == "builtins":
                 assert isinstance(annotation, type)
-                return annotation.__name__
+                return annotation.__name__, annotation.__name__
             else:
-                return self._add_import(annotation)
+                ans = self._add_import(annotation)
+                return ans, ans
         # Some sort of generic class, like a collection, or optional
         if origin == UnionType:
             # Convert new-style type unions to old-style typing.Unions
             origin = typing.Union
         if origin == Annotated:
             # If annotated, only pay attention to the first arg -- the actual type
+            if args[1] == "__chalk_primary__":
+                protocol_arg_name, init_arg_name = self._register_imports_for_annotation(args[0])
+                return f"Primary[{protocol_arg_name}]", init_arg_name
             return self._register_imports_for_annotation(args[0])
         origin_name = self._add_import(origin)
         arg_names = [self._register_imports_for_annotation(arg) for arg in args]
-        return f"{origin_name}[" + ", ".join(arg_names) + "]"
+        return (
+            f"{origin_name}[" + ", ".join(x[0] for x in arg_names) + "]",
+            f"{origin_name}[" + ", ".join(x[1] for x in arg_names) + "]",
+        )
 
     def register_features_class(self, features_cls: Type[Features]):
         annotations: Dict[str, ParsedAnnotation] = {}
@@ -129,17 +143,17 @@ class StubGenerator:
                 )
             else:
                 # Make a best-effort to figure out imports.
-                annotation = self._register_imports_for_annotation(f.typ.parsed_annotation)
+                proto_annotation, init_annotation = self._register_imports_for_annotation(f.typ.parsed_annotation)
                 include_in_proto_cls = not f.is_windowed_pseudofeature and not f.is_autogenerated
                 if include_in_proto_cls:
                     # TODO do not skip the annotations that do not appear in the protocol class
                     annotations[f.attribute_name] = ParsedAnnotation(
-                        annotation,
+                        init_annotation,
                         include_in_protocol_cls=include_in_proto_cls,
                         # If it's a features class (i.e. a has-one), not specifying
                         # the type in the protocol class so the type checker can
                         # still match feature classes that have a circular reference
-                        protocol_annotation="Any" if f.typ.is_features_cls else annotation,
+                        protocol_annotation="Any" if f.typ.is_features_cls else proto_annotation,
                     )
         if features_cls.__name__ in self._parsed_feature_classes:
             raise ValueError(
@@ -156,8 +170,11 @@ class StubGenerator:
         self._add_import("typing.Protocol", "Protocol")
         self._add_import("typing.Type", "Type")
         self._add_import("typing.overload", "overload")
+        self._add_import("typing.Iterator", "Iterator")
         self._add_import("chalk.features.feature_set.FeaturesMeta", "FeaturesMeta")
         self._add_import("chalk.features.Features", "Features")
+        self._add_import("chalk.features.Feature", "Feature")
+        self._add_import("chalk.features.primary.Primary", "Primary")
         self._add_import("chalk.utils.duration.Duration", "Duration")
         self._add_import("chalk.features.Tags", "Tags")
         lines = [
@@ -262,6 +279,11 @@ class {feature_cls_name}Meta({metaclasses_bases}):"""
     ...
 """
                 )
+            lines.append(
+                f"""\
+    def __iter__(self) -> Iterator[Feature]: ...
+"""
+            )
 
         for feature_cls_name, features_cls in parsed_feature_classes:
             # Next, generate the class definition
@@ -403,7 +425,7 @@ def configure_stubgen_argparse(parser: argparse.ArgumentParser):
     )
 
 
-def run_stubgen(args: argparse.Namespace):
+def run_stubgen(args: argparse.Namespace, file_filter: Optional[str]):
     root = args.root
     if root is None:
         root = get_directory_root() or pathlib.Path(os.getcwd())
@@ -415,7 +437,16 @@ def run_stubgen(args: argparse.Namespace):
     if stub_path is None:
         stub_path = str(root / "typings")
     assert isinstance(stub_path, str)
-    failed_imports = import_all_python_files_from_dir(root)
+    file_allowlist = None
+    if file_filter is not None and file_filter != "":
+        with open(file_filter) as file_filter_contents:
+            file_allowlist = [
+                pathlib.Path(x.strip()) for x in file_filter_contents.readlines() if x.strip().endswith(".py")
+            ]
+    failed_imports = import_all_python_files_from_dir(
+        project_root=root,
+        file_allowlist=file_allowlist,
+    )
     if len(failed_imports) > 0:
         warnings.warn(
             f"Stubs may be incomplete due to errors in loading the following files: {', '.join([x.filename for x in failed_imports])}"
